"""Test for `maas.client.viscera.machines`."""

import random
from http import HTTPStatus
from unittest.mock import Mock
from xml.etree import ElementTree

from maas.client.bones.testing.server import ApplicationBuilder
from maas.client.utils.testing import make_Credentials
from maas.client.viscera import Origin
from testtools.matchers import ContainsDict, Equals, IsInstance, MatchesStructure

from .. import machines
from ..testing import bind
from ...bones import CallError
from ...bones.testing import api_descriptions
from ...enum import NodeStatus, PowerState, PowerStopMode
from ...errors import OperationNotAllowed
from ...testing import make_name_without_spaces, TestCase
from ..pods import Pod, Pods


def make_pods_origin():
    """
    Create a new origin with Pods and Pod. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(Pods, Pod)


def make_machines_origin():
    """
    Create a new origin with Machines and Machine. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(machines.Machines, machines.Machine)


def make_get_details_coroutine(system_id, return_value=""):
    async def coroutine(system_id):
        return return_value

    return coroutine


class TestMachine(TestCase):
    def test__string_representation_includes_only_system_id_and_hostname(self):
        machine = machines.Machine(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        self.assertThat(
            repr(machine),
            Equals(
                "<Machine hostname=%(hostname)r system_id=%(system_id)r>"
                % machine._data
            ),
        )

    def test__get_power_parameters(self):
        machine = make_machines_origin().Machine(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        machine._handler.power_parameters.return_value = power_parameters
        self.assertThat(machine.get_power_parameters(), Equals(power_parameters))
        machine._handler.power_parameters.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__set_power(self):
        orig_power_type = make_name_without_spaces("power_type")
        new_power_type = make_name_without_spaces("power_type")
        machine = make_machines_origin().Machine(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
                "power_type": orig_power_type,
            }
        )
        power_parameters = {"key": make_name_without_spaces("value")}
        machine._handler.update.return_value = {"power_type": new_power_type}
        machine.set_power(new_power_type, power_parameters)
        machine._handler.update.assert_called_once_with(
            system_id=machine.system_id,
            power_type=new_power_type,
            power_parameters=power_parameters,
        )
        self.assertThat(machine.power_type, Equals(new_power_type))

    def test__abort(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }
        origin = make_machines_origin()
        machine = origin.Machine(data)
        machine._handler.abort.return_value = data
        comment = make_name_without_spaces("comment")
        self.assertThat(machine.abort(comment=comment), Equals(origin.Machine(data)))
        machine._handler.abort.assert_called_once_with(
            system_id=machine.system_id, comment=comment
        )

    def test__clear_default_gateways(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }
        origin = make_machines_origin()
        machine = origin.Machine(data)
        machine._handler.clear_default_gateways.return_value = data
        self.assertThat(machine.clear_default_gateways(), Equals(origin.Machine(data)))
        machine._handler.clear_default_gateways.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__commissioning_without_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.COMMISSIONING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.commission.return_value = data
        machine.commission()
        self.assertThat(machine.status, Equals(NodeStatus.COMMISSIONING))
        machine._handler.commission.assert_called_once_with(system_id=machine.system_id)

    def test__commissioning_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.COMMISSIONING,
        }
        ready_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.READY,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.commission.return_value = data
        machine._handler.read.return_value = ready_data
        machine.commission(wait=True, wait_interval=0.1)
        self.assertThat(machine.status, Equals(NodeStatus.READY))
        machine._handler.commission.assert_called_once_with(system_id=machine.system_id)

    def test__commissioning_and_testing_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.COMMISSIONING,
        }
        testing_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.TESTING,
        }
        ready_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.READY,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.commission.return_value = data
        machine._handler.read.side_effect = [testing_data, ready_data]
        machine.commission(wait=True, wait_interval=0.1)
        self.assertThat(machine.status, Equals(NodeStatus.READY))
        machine._handler.commission.assert_called_once_with(system_id=machine.system_id)

    def test__commission_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.COMMISSIONING,
        }
        failed_commissioning_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.FAILED_COMMISSIONING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.commission.return_value = data
        machine._handler.read.return_value = failed_commissioning_data
        self.assertRaises(
            machines.FailedCommissioning,
            machine.commission,
            wait=True,
            wait_interval=0.1,
        )

    def test__commission_with_no_tests(self):
        # Regression test for https://github.com/maas/python-libmaas/issues/185
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.COMMISSIONING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.commission.return_value = data
        machine.commission(testing_scripts=random.choice(["", [], "none"]))
        self.assertThat(machine.status, Equals(NodeStatus.COMMISSIONING))
        machine._handler.commission.assert_called_once_with(
            system_id=machine.system_id, testing_scripts=["none"]
        )

    def test__deploy_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.DEPLOYING,
        }
        deployed_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.DEPLOYED,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.deploy.return_value = data
        machine._handler.read.return_value = deployed_data
        machine.deploy(wait=True, wait_interval=0.1)
        self.assertThat(machine.status, Equals(NodeStatus.DEPLOYED))
        machine._handler.deploy.assert_called_once_with(system_id=machine.system_id)

    def test__deploy_with_kvm_install(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.READY,
        }
        deploying_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.DEPLOYING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.deploy.return_value = deploying_data
        machine.deploy(install_kvm=True, wait=False)
        machine._handler.deploy.assert_called_once_with(
            system_id=machine.system_id, install_kvm=True
        )

    def test__deploy_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.DEPLOYING,
        }
        failed_deploy_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.FAILED_DEPLOYMENT,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.deploy.return_value = data
        machine._handler.read.return_value = failed_deploy_data
        self.assertRaises(
            machines.FailedDeployment, machine.deploy, wait=True, wait_interval=0.1
        )

    def test__enter_rescue_mode(self):
        rescue_mode_machine = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
            "status": NodeStatus.ENTERING_RESCUE_MODE,
        }
        machine = make_machines_origin().Machine(rescue_mode_machine)
        machine._handler.rescue_mode.return_value = rescue_mode_machine
        self.assertThat(machine.enter_rescue_mode(), Equals(machine))
        machine._handler.rescue_mode.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__enter_rescue_mode_operation_not_allowed(self):
        machine = make_machines_origin().Machine(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        # Mock the call to content.decode in the CallError constructor
        content = Mock()
        content.decode = Mock(return_value="")
        machine._handler.rescue_mode.side_effect = CallError(
            request={"method": "GET", "uri": "www.example.com"},
            response=Mock(status=HTTPStatus.FORBIDDEN),
            content=content,
            call="",
        )
        self.assertRaises(OperationNotAllowed, machine.enter_rescue_mode)

    def test__enter_rescue_mode_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        rm_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.ENTERING_RESCUE_MODE,
        }
        erm_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.RESCUE_MODE,
        }
        rm_machine = make_machines_origin().Machine(rm_data)
        erm_machine = make_machines_origin().Machine(erm_data)
        rm_machine._handler.rescue_mode.return_value = rm_data
        rm_machine._handler.read.return_value = erm_data
        result = rm_machine.enter_rescue_mode(wait=True, wait_interval=0.1)
        self.assertThat(result.status, Equals(erm_machine.status))
        rm_machine._handler.rescue_mode.assert_called_once_with(
            system_id=rm_machine.system_id
        )

    def test__enter_rescue_mode_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.ENTERING_RESCUE_MODE,
        }
        failed_enter_rescue_mode_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.FAILED_ENTERING_RESCUE_MODE,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.rescue_mode.return_value = data
        machine._handler.read.return_value = failed_enter_rescue_mode_data
        self.assertRaises(
            machines.RescueModeFailure,
            machine.enter_rescue_mode,
            wait=True,
            wait_interval=0.1,
        )

    def test__exit_rescue_mode(self):
        exit_machine = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
            "status": NodeStatus.EXITING_RESCUE_MODE,
        }
        machine = make_machines_origin().Machine(exit_machine)
        machine._handler.exit_rescue_mode.return_value = exit_machine
        self.assertThat(machine.exit_rescue_mode(), Equals(machine))
        machine._handler.exit_rescue_mode.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__exit_rescue_mode_operation_not_allowed(self):
        machine = make_machines_origin().Machine(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        # Mock the call to content.decode in the CallError constructor
        content = Mock()
        content.decode = Mock(return_value="")
        machine._handler.exit_rescue_mode.side_effect = CallError(
            request={"method": "GET", "uri": "www.example.com"},
            response=Mock(status=HTTPStatus.FORBIDDEN),
            content=content,
            call="",
        )
        self.assertRaises(OperationNotAllowed, machine.exit_rescue_mode)

    def test__exit_rescue_mode_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.EXITING_RESCUE_MODE,
        }
        deployed_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.DEPLOYED,
        }
        machine = make_machines_origin().Machine(data)
        deployed_machine = make_machines_origin().Machine(deployed_data)
        machine._handler.exit_rescue_mode.return_value = data
        machine._handler.read.return_value = deployed_data
        result = machine.exit_rescue_mode(wait=True, wait_interval=0.1)
        self.assertThat(result.status, Equals(deployed_machine.status))
        machine._handler.exit_rescue_mode.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__exit_rescue_mode_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.EXITING_RESCUE_MODE,
        }
        failed_exit_rescue_mode_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.FAILED_EXITING_RESCUE_MODE,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.exit_rescue_mode.return_value = data
        machine._handler.read.return_value = failed_exit_rescue_mode_data
        self.assertRaises(
            machines.RescueModeFailure,
            machine.exit_rescue_mode,
            wait=True,
            wait_interval=0.1,
        )

    def test__get_curtin_config(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }
        origin = make_machines_origin()
        machine = origin.Machine(data)
        config = make_name_without_spaces("config")
        machine._handler.get_curtin_config.return_value = config
        self.assertThat(machine.get_curtin_config(), Equals(config))
        machine._handler.get_curtin_config.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__mark_broken(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }
        origin = make_machines_origin()
        machine = origin.Machine(data)
        machine._handler.mark_broken.return_value = data
        comment = make_name_without_spaces("comment")
        self.assertThat(
            machine.mark_broken(comment=comment), Equals(origin.Machine(data))
        )
        machine._handler.mark_broken.assert_called_once_with(
            system_id=machine.system_id, comment=comment
        )

    def test__mark_fixed(self):
        data = {
            "system_id": make_name_without_spaces("system-id"),
            "hostname": make_name_without_spaces("hostname"),
        }
        origin = make_machines_origin()
        machine = origin.Machine(data)
        machine._handler.mark_fixed.return_value = data
        comment = make_name_without_spaces("comment")
        self.assertThat(
            machine.mark_fixed(comment=comment), Equals(origin.Machine(data))
        )
        machine._handler.mark_fixed.assert_called_once_with(
            system_id=machine.system_id, comment=comment
        )

    def test__release_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.RELEASING,
        }
        allocated_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.ALLOCATED,
        }
        machine = make_machines_origin().Machine(data)
        allocated_machine = make_machines_origin().Machine(allocated_data)
        machine._handler.release.return_value = data
        machine._handler.read.return_value = allocated_data
        result = machine.release(wait=True, wait_interval=0.1)
        self.assertThat(result.status, Equals(allocated_machine.status))
        machine._handler.release.assert_called_once_with(system_id=machine.system_id)

    def test__release_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.RELEASING,
        }
        failed_release_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.FAILED_RELEASING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.release.return_value = data
        machine._handler.read.return_value = failed_release_data
        self.assertRaises(
            machines.FailedReleasing, machine.release, wait=True, wait_interval=0.1
        )

    def test__release_with_wait_failed_disk_erasing(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.RELEASING,
        }
        failed_disk_erase_data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.FAILED_DISK_ERASING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.release.return_value = data
        machine._handler.read.return_value = failed_disk_erase_data
        self.assertRaises(
            machines.FailedDiskErasing, machine.release, wait=True, wait_interval=0.1
        )

    def test__release_with_deleted_machine(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "status": NodeStatus.RELEASING,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.release.return_value = data
        machine._handler.read.side_effect = CallError(
            request={"method": "GET", "uri": "www.example.com"},
            response=Mock(status=HTTPStatus.NOT_FOUND),
            content=b"",
            call="",
        )
        result = machine.release(wait=True, wait_interval=0.1)
        self.assertThat(result.status, Equals(NodeStatus.RELEASING))
        machine._handler.release.assert_called_once_with(system_id=machine.system_id)

    def test__power_on_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.OFF,
        }
        power_data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.ON,
        }
        machine = make_machines_origin().Machine(data)
        powered_machine = make_machines_origin().Machine(power_data)
        machine._handler.power_on.return_value = data
        machine._handler.read.return_value = power_data
        result = machine.power_on(wait=True, wait_interval=0.1)
        self.assertThat(result.power_state, Equals(powered_machine.power_state))
        machine._handler.power_on.assert_called_once_with(system_id=machine.system_id)

    def test__power_on_doesnt_wait_for_unknown(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.UNKNOWN,
        }
        power_data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.UNKNOWN,
        }
        machine = make_machines_origin().Machine(data)
        powered_machine = make_machines_origin().Machine(power_data)
        machine._handler.power_on.return_value = data
        machine._handler.read.return_value = power_data
        result = machine.power_on(wait=True, wait_interval=0.1)
        self.assertThat(result.power_state, Equals(powered_machine.power_state))
        assert machine._handler.read.call_count == 0

    def test__power_on_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.OFF,
        }
        failed_power_data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.ERROR,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.power_on.return_value = data
        machine._handler.read.return_value = failed_power_data
        self.assertRaises(
            machines.PowerError, machine.power_on, wait=True, wait_interval=0.1
        )

    def test__power_off_with_wait(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.ON,
        }
        power_data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.OFF,
        }
        machine = make_machines_origin().Machine(data)
        powered_machine = make_machines_origin().Machine(power_data)
        machine._handler.power_off.return_value = data
        machine._handler.read.return_value = power_data
        result = machine.power_off(wait=True, wait_interval=0.1)
        self.assertThat(result.power_state, Equals(powered_machine.power_state))
        machine._handler.power_off.assert_called_once_with(
            system_id=machine.system_id, stop_mode=PowerStopMode.HARD.value
        )

    def test__power_off_soft_mode(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.ON,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.power_off.return_value = data
        machine.power_off(stop_mode=PowerStopMode.SOFT, wait=False)
        machine._handler.power_off.assert_called_once_with(
            system_id=machine.system_id, stop_mode=PowerStopMode.SOFT.value
        )

    def test__power_off_doesnt_wait_for_unknown(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.UNKNOWN,
        }
        power_data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.UNKNOWN,
        }
        machine = make_machines_origin().Machine(data)
        powered_machine = make_machines_origin().Machine(power_data)
        machine._handler.power_off.return_value = data
        machine._handler.read.return_value = power_data
        result = machine.power_off(wait=True, wait_interval=0.1)
        self.assertThat(result.power_state, Equals(powered_machine.power_state))
        self.assertThat(machine._handler.read.call_count, Equals(0))

    def test__power_off_with_wait_failed(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.ON,
        }
        failed_power_data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.ERROR,
        }
        machine = make_machines_origin().Machine(data)
        machine._handler.power_off.return_value = data
        machine._handler.read.return_value = failed_power_data
        self.assertRaises(
            machines.PowerError, machine.power_off, wait=True, wait_interval=0.1
        )

    def test__query_power_state(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "power_state": PowerState.OFF,
        }
        query_data = {"state": "on"}
        machine = make_machines_origin().Machine(data)
        machine._handler.query_power_state.return_value = query_data
        result = machine.query_power_state()
        self.assertIsInstance(result, PowerState)
        self.assertEquals(PowerState.ON, result)
        self.assertEqual(PowerState.ON, machine.power_state)

    def test__get_details(self):
        return_val = (
            b"S\x01\x00\x00\x05lshw\x00\xf2\x00\x00\x00\x00<?xml"
            b' version="1.0" standalone="yes" ?>\n<!-- generated by'
            b" lshw-B.02.17 -->\n<!-- GCC 5.3.1 20160413 -->\n<!--"
            b" Linux 4.8.0-34-generic #36~16.04.1-Ubuntu SMP Wed"
            b" Dec 21 18:55:08 UTC 2016 x86_64 -->\n<!-- GNU libc 2"
            b" (glibc 2.23) -->\n<list>\n</list>\n\x05lldp\x00F\x00"
            b'\x00\x00\x00<?xml version="1.0" encoding="UTF-8"?>\n'
            b'<lldp label="LLDP neighbors"/>\n\x00'
        )
        machine = make_machines_origin().Machine(
            {
                "system_id": make_name_without_spaces("system-id"),
                "hostname": make_name_without_spaces("hostname"),
            }
        )
        machine._handler.details = make_get_details_coroutine(
            machine.system_id, return_value=return_val
        )
        data = machine.get_details()
        self.assertItemsEqual(["lldp", "lshw"], data.keys())
        lldp = ElementTree.fromstring(data["lldp"])
        lshw = ElementTree.fromstring(data["lshw"])
        assert IsInstance(lldp, ElementTree)
        assert IsInstance(lshw, ElementTree)

    def test__restore_default_configuration(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {"system_id": system_id, "hostname": hostname}
        machine = make_machines_origin().Machine(data)
        machine._handler.restore_default_configuration.return_value = data
        machine.restore_default_configuration()
        self.assertEqual(data, machine._data)
        machine._handler.restore_default_configuration.assert_called_once_with(
            system_id=machine.system_id
        )

    def test__restore_networking_configuration(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {"system_id": system_id, "hostname": hostname}
        machine = make_machines_origin().Machine(data)
        mock_restore_networking = machine._handler.restore_networking_configuration
        mock_restore_networking.return_value = data
        machine.restore_networking_configuration()
        self.assertEqual(data, machine._data)
        mock_restore_networking.assert_called_once_with(system_id=machine.system_id)

    def test__restore_storage_configuration(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {"system_id": system_id, "hostname": hostname}
        machine = make_machines_origin().Machine(data)
        mock_restore_storage = machine._handler.restore_storage_configuration
        mock_restore_storage.return_value = data
        machine.restore_storage_configuration()
        self.assertEqual(data, machine._data)
        mock_restore_storage.assert_called_once_with(system_id=machine.system_id)

    def test__save_updates_owner_data(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "owner_data": {"hello": "world", "delete": "me", "keep": "me"},
        }
        machine = make_machines_origin().Machine(data)
        del machine.owner_data["delete"]
        machine.owner_data["hello"] = "whole new world"
        machine.owner_data["new"] = "brand-new"
        machine._handler.set_owner_data.return_value = {}
        machine.save()
        self.assertThat(machine._handler.update.call_count, Equals(0))
        self.assertThat(
            machine.owner_data,
            Equals({"hello": "whole new world", "new": "brand-new", "keep": "me"}),
        )
        machine._handler.set_owner_data.assert_called_once_with(
            system_id=machine.system_id,
            delete="",
            hello="whole new world",
            new="brand-new",
        )

    def test__save_updates_owner_data_with_replace(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {
            "system_id": system_id,
            "hostname": hostname,
            "owner_data": {"hello": "world", "delete": "me", "keep": "me"},
        }
        machine = make_machines_origin().Machine(data)
        machine.owner_data = {
            "hello": "whole new world",
            "keep": "me",
            "new": "brand-new",
        }
        machine._handler.set_owner_data.return_value = {}
        machine.save()
        self.assertThat(machine._handler.update.call_count, Equals(0))
        self.assertThat(
            machine.owner_data,
            Equals({"hello": "whole new world", "new": "brand-new", "keep": "me"}),
        )
        machine._handler.set_owner_data.assert_called_once_with(
            system_id=machine.system_id,
            delete="",
            hello="whole new world",
            new="brand-new",
        )

    def test__lock(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {"system_id": system_id, "hostname": hostname, "locked": False}
        new_data = {"system_id": system_id, "hostname": hostname, "locked": True}
        origin = make_machines_origin()
        machine = origin.Machine(data)
        machine._handler.lock.return_value = new_data
        comment = make_name_without_spaces("comment")
        self.assertThat(machine.lock(comment=comment), Equals(origin.Machine(new_data)))
        machine._handler.lock.assert_called_once_with(
            system_id=machine.system_id, comment=comment
        )
        self.assertThat(machine.locked, Equals(new_data["locked"]))

    def test__unlock(self):
        system_id = make_name_without_spaces("system-id")
        hostname = make_name_without_spaces("hostname")
        data = {"system_id": system_id, "hostname": hostname, "locked": True}
        new_data = {"system_id": system_id, "hostname": hostname, "locked": False}
        origin = make_machines_origin()
        machine = origin.Machine(data)
        machine._handler.unlock.return_value = new_data
        comment = make_name_without_spaces("comment")
        self.assertThat(
            machine.unlock(comment=comment), Equals(origin.Machine(new_data))
        )
        machine._handler.unlock.assert_called_once_with(
            system_id=machine.system_id, comment=comment
        )
        self.assertThat(machine.locked, Equals(new_data["locked"]))


class TestMachine_APIVersion(TestCase):

    scenarios = tuple(
        (name, dict(version=version, description=description))
        for name, version, description in api_descriptions
    )

    async def test__deploy(self):
        builder = ApplicationBuilder(self.description)

        @builder.handle("auth:Machine.deploy")
        async def deploy(request, system_id):
            self.assertThat(
                request.params,
                ContainsDict(
                    {
                        "distro_series": Equals("ubuntu/xenial"),
                        "hwe_kernel": Equals("hwe-x"),
                    }
                ),
            )
            return {"system_id": system_id, "status_name": "Deploying"}

        async with builder.serve() as baseurl:
            origin = await Origin.fromURL(baseurl, credentials=make_Credentials())
            machine = origin.Machine(
                {"system_id": make_name_without_spaces("system-id")}
            )
            machine = await machine.deploy(
                distro_series="ubuntu/xenial", hwe_kernel="hwe-x"
            )
            self.assertThat(
                machine, MatchesStructure.byEquality(status_name="Deploying")
            )


class TestMachines(TestCase):
    def test__create(self):
        origin = make_machines_origin()
        Machines, Machine = origin.Machines, origin.Machine
        Machines._handler.create.return_value = {}
        observed = Machines.create(
            "amd64",
            ["00:11:22:33:44:55", "00:11:22:33:44:AA"],
            "ipmi",
            {"power_address": "localhost", "power_user": "root"},
            subarchitecture="generic",
            min_hwe_kernel="hwe-x",
            hostname="new-machine",
            domain="maas",
        )
        self.assertThat(observed, IsInstance(Machine))
        Machines._handler.create.assert_called_once_with(
            architecture="amd64",
            mac_addresses=["00:11:22:33:44:55", "00:11:22:33:44:AA"],
            power_type="ipmi",
            power_parameters=(
                '{"power_address": "localhost", ' '"power_user": "root"}'
            ),
            subarchitecture="generic",
            min_hwe_kernel="hwe-x",
            hostname="new-machine",
            domain="maas",
        )

    def test__allocate(self):
        Machines = make_machines_origin().Machines
        Machines._handler.allocate.return_value = {}
        hostname = make_name_without_spaces("hostname")
        Machines.allocate(
            hostname=hostname,
            architectures=["amd64/generic"],
            cpus=4,
            memory=1024.0,
            tags=["foo", "bar"],
            not_tags=["baz"],
        )
        Machines._handler.allocate.assert_called_once_with(
            name=hostname,  # API parameter is actually name, not hostname
            arch=["amd64/generic"],
            cpu_count="4",
            mem="1024.0",
            tags=["foo", "bar"],
            not_tags=["baz"],
        )

    def test__allocate_with_pod(self):
        Pod = make_pods_origin().Pod
        pod = Pod({"name": make_name_without_spaces("pod")})
        Machines = make_machines_origin().Machines
        Machines._handler.allocate.return_value = {}
        hostname = make_name_without_spaces("hostname")
        Machines.allocate(
            hostname=hostname,
            architectures=["amd64/generic"],
            cpus=4,
            memory=1024.0,
            tags=["foo", "bar"],
            not_tags=["baz"],
            pod=pod.name,
        )
        Machines._handler.allocate.assert_called_once_with(
            name=hostname,  # API parameter is actually name, not hostname
            arch=["amd64/generic"],
            cpu_count="4",
            mem="1024.0",
            tags=["foo", "bar"],
            not_tags=["baz"],
            pod=pod.name,
        )

    def test__allocate_with_not_pod(self):
        Pod = make_pods_origin().Pod
        pod = Pod({"name": make_name_without_spaces("pod")})
        Machines = make_machines_origin().Machines
        Machines._handler.allocate.return_value = {}
        hostname = make_name_without_spaces("hostname")
        Machines.allocate(
            hostname=hostname,
            architectures=["amd64/generic"],
            cpus=4,
            memory=1024.0,
            tags=["foo", "bar"],
            not_tags=["baz"],
            not_pod=pod.name,
        )
        Machines._handler.allocate.assert_called_once_with(
            name=hostname,  # API parameter is actually name, not hostname
            arch=["amd64/generic"],
            cpu_count="4",
            mem="1024.0",
            tags=["foo", "bar"],
            not_tags=["baz"],
            not_pod=pod.name,
        )

    def test__get_power_parameters_for_with_empty_list(self):
        Machines = make_machines_origin().Machines
        self.assertThat(Machines.get_power_parameters_for(system_ids=[]), Equals({}))

    def test__get_power_parameters_for_with_system_ids(self):
        power_parameters = {
            make_name_without_spaces("system_id"): {
                "key": make_name_without_spaces("value")
            }
            for _ in range(3)
        }
        Machines = make_machines_origin().Machines
        Machines._handler.power_parameters.return_value = power_parameters
        self.assertThat(
            Machines.get_power_parameters_for(system_ids=power_parameters.keys()),
            Equals(power_parameters),
        )
        Machines._handler.power_parameters.assert_called_once_with(
            id=power_parameters.keys()
        )



"""Objects for boot sources."""

__all__ = ["BootSource", "BootSources"]

from . import check, Object, ObjectField, ObjectSet, ObjectType, parse_timestamp
from ..utils import coalesce


class BootSourcesType(ObjectType):
    """Metaclass for `BootSources`."""

    async def create(cls, url, *, keyring_filename=None, keyring_data=None):
        """Create a new `BootSource`.

        :param url: The URL for the boot source.
        :param keyring_filename: The path to the keyring file on the server.
        :param keyring_data: The GPG keyring data, binary. as a file-like
            object. For example: an open file handle in binary mode, or an
            instance of `io.BytesIO`.
        """
        data = await cls._handler.create(
            url=url,
            keyring_filename=coalesce(keyring_filename, ""),
            keyring_data=coalesce(keyring_data, ""),
        )
        return cls._object(data)

    async def read(cls):
        """Get list of `BootSource`'s."""
        data = await cls._handler.read()
        return cls(map(cls._object, data))


class BootSources(ObjectSet, metaclass=BootSourcesType):
    """The set of boot sources."""


class BootSourceType(ObjectType):
    async def read(cls, id):
        """Get `BootSource` by `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class BootSource(Object, metaclass=BootSourceType):
    """A boot source."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    url = ObjectField.Checked("url", check(str), check(str))
    keyring_filename = ObjectField.Checked(
        "keyring_filename", check(str), check(str), default=""
    )
    keyring_data = ObjectField.Checked(
        "keyring_data", check(str), check(str), default=""
    )
    created = ObjectField.Checked("created", parse_timestamp, readonly=True)
    updated = ObjectField.Checked("updated", parse_timestamp, readonly=True)

    def __repr__(self):
        return super(BootSource, self).__repr__(
            fields={"url", "keyring_filename", "keyring_data"}
        )

    async def delete(self):
        """Delete boot source."""
        await self._handler.delete(id=self.id)



"""Abstractions around API description documents."""

__all__ = ["Description", "Action"]

from keyword import iskeyword
from operator import itemgetter

from maas.client.bones.helpers import derive_resource_name
from maas.client.utils import parse_docstring


class Description:
    """Object-oriented interface to a MAAS API description document."""

    def __init__(self, description):
        super(Description, self).__init__()
        self._description = description
        self._populate()

    def _populate(self):
        self.anon = Resources(True, self._resources("anon"))
        self.auth = Resources(False, self._resources("auth"))

    def _resources(self, classification):
        resources = self._description["resources"]
        resources = map(itemgetter(classification), resources)
        return (rs for rs in resources if rs is not None)

    @property
    def doc(self):
        doc = self._description["doc"]
        return parse_docstring(doc)

    @property
    def hash(self):
        return self._description["hash"]

    @property
    def raw(self):
        return self._description

    def __iter__(self):
        """Iterate all resources, anonymous first."""
        yield from self.anon
        yield from self.auth

    def __repr__(self):
        title, body = self.doc
        return "<%s %r %s>" % (self.__class__.__name__, title.rstrip("."), self.hash)


class Resources:
    """Pincushion of API resources."""

    def __init__(self, is_anonymous, resources):
        super(Resources, self).__init__()
        for resource in resources:
            name = derive_resource_name(resource["name"])
            resource = Resource(name, is_anonymous, resource)
            attrname = "%s_" % name if iskeyword(name) else name
            setattr(self, attrname, resource)

    def __iter__(self):
        """Iterate all resources."""
        for value in vars(self).values():
            if isinstance(value, Resource):
                yield value


class Resource:
    """An API resource, like `Machines`."""

    def __init__(self, name, is_anonymous, data):
        super(Resource, self).__init__()
        self._is_anonymous = is_anonymous
        self._name = name
        self._data = data
        self._populate()

    def _populate(self):
        for action in self._data["actions"]:
            name = action["name"]
            name = "%s_" % name if iskeyword(name) else name
            setattr(self, name, Action(self, action))
        self._properties = {
            "doc": parse_docstring(self._data["doc"]),
            "is_anonymous": self._is_anonymous,
            "name": self._name,
            "name/raw": self._data["name"],
            "params": tuple(self._data["params"]),
            "path": self._data["path"],
            "uri": self._data["uri"],
        }

    def __getitem__(self, name):
        return self._properties[name]

    def __iter__(self):
        """Iterate all actions."""
        for value in vars(self).values():
            if isinstance(value, Action):
                yield value

    def __repr__(self):
        title, body = self["doc"]
        return "<%s:%s %r>" % (self.__class__.__name__, self._name, title.rstrip("."))


class Action:
    """An API action on a resource, like `Machines.allocate`."""

    def __init__(self, resource, data):
        super(Action, self).__init__()
        self._resource = resource
        self._data = data

    # Resource-specific properties.

    @property
    def resource(self):
        return self._resource

    @property
    def is_anonymous(self):
        return self._resource["is_anonymous"]

    @property
    def params(self):
        return self._resource["params"]

    @property
    def path(self):
        return self._resource["path"]

    @property
    def uri(self):
        return self._resource["uri"]

    # Action-specific properties.

    @property
    def doc(self):
        doc = self._data["doc"]
        return parse_docstring(doc)

    @property
    def method(self):
        return self._data["method"]

    @property
    def name(self):
        return self._data["name"]

    @property
    def op(self):
        return self._data["op"]

    @property
    def is_restful(self):
        return self._data["restful"]

    @property
    def action_name(self):
        anon_auth = "anon" if self.is_anonymous else "auth"
        return "%s:%s.%s" % (anon_auth, self.resource["name"], self.name)

    # Other.

    def __repr__(self):
        title, body = self.doc
        return "<%s:%s.%s %r %s %s%s>" % (
            self.__class__.__name__,
            self._resource._name,
            self.name,
            title.rstrip("."),
            self.method,
            self.path,
            ("" if self.op is None else "?op=" + self.op),
        )



"""Objects for static_routes."""

__all__ = ["StaticRoutes", "StaticRoute"]

from . import check, Object, ObjectField, ObjectFieldRelated, ObjectSet, ObjectType
from .subnets import Subnet
from typing import Union


class StaticRoutesType(ObjectType):
    """Metaclass for `StaticRoutes`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(
        cls,
        destination: Union[int, Subnet],
        source: Union[int, Subnet],
        gateway_ip: str,
        metric: int,
    ):
        """
        Create a `StaticRoute` in MAAS.

        :param name: The name of the `StaticRoute` (optional, will be given a
        default value if not specified).
        :type name: `str`
        :param description: A description of the `StaticRoute` (optional).
        :type description: `str`
        :param class_type: The class type of the `StaticRoute` (optional).
        :type class_type: `str`
        :returns: The created StaticRoute
        :rtype: `StaticRoute`
        """
        params = {"gateway_ip": gateway_ip, "metric": metric}
        if isinstance(source, Subnet):
            params["source"] = source.id
        elif isinstance(source, int):
            params["source"] = source
        if isinstance(destination, Subnet):
            params["destination"] = destination.id
        elif isinstance(destination, int):
            params["destination"] = destination
        return cls._object(await cls._handler.create(**params))


class StaticRoutes(ObjectSet, metaclass=StaticRoutesType):
    """The set of StaticRoutes stored in MAAS."""


class StaticRouteType(ObjectType):
    """Metaclass for `StaticRoute`."""

    async def read(cls, id: int):
        """Get a `StaticRoute` by its `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class StaticRoute(Object, metaclass=StaticRouteType):
    """A StaticRoute."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    destination = ObjectFieldRelated("destination", "Subnet")
    source = ObjectFieldRelated("source", "Subnet")
    gateway_ip = ObjectField.Checked("gateway_ip", check(str))
    metric = ObjectField.Checked("metric", check(int))

    async def delete(self):
        """Delete this StaticRoute."""
        await self._handler.delete(id=self.id)



"""Miscellaneous helpers for Bones."""

__all__ = [
    "authenticate",
    "connect",
    "ConnectError",
    "derive_resource_name",
    "fetch_api_description",
    "login",
    "LoginError",
    "LoginNotSupported",
    "PasswordWithoutUsername",
    "RemoteError",
    "UsernameWithoutPassword",
]

import asyncio
from concurrent import futures
from getpass import getuser
from http import HTTPStatus
from socket import gethostname
import typing
from urllib.parse import ParseResult, SplitResult, urljoin, urlparse

import aiohttp
from macaroonbakery import httpbakery

from ..utils import api_url
from ..utils.maas_async import asynchronous
from ..utils.creds import Credentials
from ..utils.profiles import Profile


class RemoteError(Exception):
    """Miscellaneous error related to a remote system."""


async def fetch_api_description(
    url: typing.Union[str, ParseResult, SplitResult], insecure: bool = False
):
    """Fetch the API description from the remote MAAS instance."""
    url_describe = urljoin(_ensure_url_string(url), "describe/")
    connector = aiohttp.TCPConnector(verify_ssl=(not insecure))
    session = aiohttp.ClientSession(connector=connector)
    async with session, session.get(url_describe) as response:
        if response.status != HTTPStatus.OK:
            raise RemoteError("{0} -> {1.status} {1.reason}".format(url, response))
        elif response.content_type != "application/json":
            raise RemoteError(
                "Expected application/json, got: %s" % response.content_type
            )
        else:
            return await response.json()


def _ensure_url_string(url):
    """Convert `url` to a string URL if it isn't one already."""
    if isinstance(url, str):
        return url
    elif isinstance(url, (ParseResult, SplitResult)):
        return url.geturl()
    else:
        raise TypeError("Could not convert %r to a string URL." % (url,))


def derive_resource_name(name):
    """A stable, human-readable name and identifier for a resource."""
    if name.startswith("Anon"):
        name = name[4:]
    if name.endswith("Handler"):
        name = name[:-7]
    if name == "Maas":
        name = "MAAS"
    return name


class ConnectError(Exception):
    """An error with connecting."""


@asynchronous
async def connect(url, *, apikey=None, insecure=False):
    """Connect to a remote MAAS instance with `apikey`.

    Returns a new :class:`Profile` which has NOT been saved. To connect AND
    save a new profile::

        profile = connect(url, apikey=apikey)
        profile = profile.replace(name="mad-hatter")

        with profiles.ProfileStore.open() as config:
            config.save(profile)
            # Optionally, set it as the default.
            config.default = profile.name

    """
    url = api_url(url)
    url = urlparse(url)

    if url.username is not None:
        raise ConnectError(
            "Cannot provide user-name explicitly in URL (%r) when connecting; "
            "use login instead." % url.username
        )
    if url.password is not None:
        raise ConnectError(
            "Cannot provide password explicitly in URL (%r) when connecting; "
            "use login instead." % url.username
        )

    if apikey is None:
        credentials = None  # Anonymous access.
    else:
        credentials = Credentials.parse(apikey)

    description = await fetch_api_description(url, insecure)

    # Return a new (unsaved) profile.
    return Profile(
        name=url.netloc,
        url=url.geturl(),
        credentials=credentials,
        description=description,
        insecure=insecure,
    )


class LoginError(Exception):
    """An error with logging-in."""


class PasswordWithoutUsername(LoginError):
    """A password was provided without a corresponding user-name."""


class UsernameWithoutPassword(LoginError):
    """A user-name was provided without a corresponding password."""


class LoginNotSupported(LoginError):
    """Server does not support login-type auth for API clients."""


class MacaroonLoginNotSupported(LoginError):
    """Server does not support macaroon auth for API clients."""


@asynchronous
async def login(url, *, anonymous=False, username=None, password=None, insecure=False):
    """Log-in to a remote MAAS instance.

    Returns a new :class:`Profile` which has NOT been saved. To log-in AND
    save a new profile::

        profile = login(url, username="alice", password="wonderland")
        profile = profile.replace(name="mad-hatter")

        with profiles.ProfileStore.open() as config:
            config.save(profile)
            # Optionally, set it as the default.
            config.default = profile.name

    :raise RemoteError: An unexpected error from the remote system.
    :raise LoginError: An error related to logging-in.
    :raise PasswordWithoutUsername: Password given, but not username.
    :raise UsernameWithoutPassword: Username given, but not password.
    :raise LoginNotSupported: Server does not support API client log-in.
    """
    url = api_url(url)
    url = urlparse(url)

    if username is None:
        username = url.username
    else:
        if url.username is None:
            pass  # Anonymous access.
        else:
            raise LoginError(
                "User-name provided explicitly (%r) and in URL (%r); "
                "provide only one." % (username, url.username)
            )

    if password is None:
        password = url.password
    else:
        if url.password is None:
            pass  # Anonymous access.
        else:
            raise LoginError(
                "Password provided explicitly (%r) and in URL (%r); "
                "provide only one." % (password, url.password)
            )

    # Remove user-name and password from the URL.
    userinfo, _, hostinfo = url.netloc.rpartition("@")
    url = url._replace(netloc=hostinfo)

    if username is None:
        if password:
            raise PasswordWithoutUsername(
                "Password provided without user-name; specify user-name."
            )
        elif anonymous:
            credentials = None
        else:
            credentials = await authenticate_with_macaroon(
                url.geturl(), insecure=insecure
            )
    else:
        if password is None:
            raise UsernameWithoutPassword(
                "User-name provided without password; specify password."
            )
        else:
            credentials = await authenticate(
                url.geturl(), username, password, insecure=insecure
            )

    description = await fetch_api_description(url, insecure)
    profile_name = username or url.netloc

    # Return a new (unsaved) profile.
    return Profile(
        name=profile_name,
        url=url.geturl(),
        credentials=credentials,
        description=description,
        insecure=insecure,
    )


async def authenticate_with_macaroon(url, insecure=False):
    """Login via macaroons and generate and return new API keys."""
    executor = futures.ThreadPoolExecutor(max_workers=1)

    def get_token():
        client = httpbakery.Client()
        resp = client.request(
            "POST",
            "{}/account/?op=create_authorisation_token".format(url),
            verify=not insecure,
        )
        if resp.status_code == HTTPStatus.UNAUTHORIZED:
            # if the auteentication with Candid fails, an exception is raised
            # above so we don't get here
            raise MacaroonLoginNotSupported("Macaroon authentication not supported")
        if resp.status_code != HTTPStatus.OK:
            raise LoginError("Login failed: {}".format(resp.text))
        result = resp.json()
        return "{consumer_key}:{token_key}:{token_secret}".format(**result)

    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(executor, get_token)


async def authenticate(url, username, password, *, insecure=False):
    """Obtain a new API key by logging into MAAS.

    :param url: URL for the MAAS API (i.e. ends with ``/api/x.y/``).
    :param insecure: If true, don't verify SSL/TLS certificates.
    :return: A `Credentials` instance.

    :raise RemoteError: An unexpected error from the remote system.
    :raise LoginNotSupported: Server does not support API client log-in.
    """
    url_versn = urljoin(url, "version/")
    url_authn = urljoin(url, "../../accounts/authenticate/")

    def check_response_is_okay(response):
        if response.status != HTTPStatus.OK:
            raise RemoteError(
                "{0} -> {1.status} {1.reason}".format(
                    response.url_obj.human_repr(), response
                )
            )

    connector = aiohttp.TCPConnector(verify_ssl=(not insecure))
    session = aiohttp.ClientSession(connector=connector)
    async with session:
        # Check that this server supports `authenticate-api`.
        async with session.get(url_versn) as response:
            check_response_is_okay(response)
            version_info = await response.json()

        if "authenticate-api" not in version_info["capabilities"]:
            raise LoginNotSupported(
                "Server does not support automated client log-in. "
                "Please obtain an API token via the MAAS UI."
            )

        # POST to the `authenticate` endpoint.
        data = {
            "username": username,
            "password": password,
            "consumer": "%s@%s" % (getuser(), gethostname()),
        }
        async with session.post(url_authn, data=data) as response:
            check_response_is_okay(response)
            token_info = await response.json()

        return Credentials(
            token_info["consumer_key"],
            token_info["token_key"],
            token_info["token_secret"],
        )



"""Tests for `maas.client.viscera.zones`."""

import random

from testtools.matchers import Equals, IsInstance, MatchesStructure

from .. import zones

from ..testing import bind
from ...testing import make_name, make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Zone and Zones. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(zones.Zones, zones.Zone)


class TestZones(TestCase):
    def test__zones_create(self):
        origin = make_origin()
        zone_id = random.randint(0, 100)
        name = make_string_without_spaces()
        description = make_string_without_spaces()
        origin.Zones._handler.create.return_value = {
            "id": zone_id,
            "name": name,
            "description": description,
        }
        zone = origin.Zones.create(name=name, description=description)
        origin.Zones._handler.create.assert_called_once_with(
            name=name, description=description
        )
        self.assertThat(zone, IsInstance(origin.Zone))
        self.assertThat(
            zone,
            MatchesStructure.byEquality(id=zone_id, name=name, description=description),
        )

    def test__zones_create_without_description(self):
        origin = make_origin()
        zone_id = random.randint(0, 100)
        name = make_string_without_spaces()
        description = ""
        origin.Zones._handler.create.return_value = {
            "id": zone_id,
            "name": name,
            "description": description,
        }
        zone = origin.Zones.create(name=name)
        origin.Zones._handler.create.assert_called_once_with(name=name)
        self.assertThat(zone, IsInstance(origin.Zone))
        self.assertThat(
            zone,
            MatchesStructure.byEquality(id=zone_id, name=name, description=description),
        )

    def test__zones_read(self):
        """Zones.read() returns a list of Zones."""
        Zones = make_origin().Zones
        zones = [
            {
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "description": make_string_without_spaces(),
            }
            for _ in range(3)
        ]
        Zones._handler.read.return_value = zones
        zones = Zones.read()
        self.assertThat(len(zones), Equals(3))


class TestZone(TestCase):
    def test__zone_unloaded(self):
        Zone = make_origin().Zone
        name = make_name("zone")
        zone = Zone(name)
        self.assertFalse(zone.loaded)
        self.assertEqual(name, zone.name)

    def test__zone_read(self):
        Zone = make_origin().Zone
        zone = {
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "description": make_string_without_spaces(),
        }
        Zone._handler.read.return_value = zone
        self.assertThat(Zone.read(name=zone["name"]), Equals(Zone(zone)))
        Zone._handler.read.assert_called_once_with(name=zone["name"])

    def test__zone_delete(self):
        Zone = make_origin().Zone
        zone_name = make_string_without_spaces()
        zone = Zone(
            {
                "id": random.randint(0, 100),
                "name": zone_name,
                "description": make_string_without_spaces(),
            }
        )
        zone.delete()
        Zone._handler.delete.assert_called_once_with(name=zone_name)



"""Objects for filesystems."""

__all__ = ["Filesystem"]

from . import check, Object, ObjectField


class Filesystem(Object):
    """A filesystem on either a partition or block device."""

    label = ObjectField.Checked("label", check(str), readonly=True)
    fstype = ObjectField.Checked("fstype", check(str), readonly=True)
    mount_point = ObjectField.Checked("mount_point", check(str), readonly=True)
    mount_options = ObjectField.Checked("mount_options", check(str), readonly=True)
    uuid = ObjectField.Checked("uuid", check(str), readonly=True)

    def __repr__(self):
        return super(Filesystem, self).__repr__(fields={"fstype", "mount_point"})



"""Test for `maas.client.viscera.users`."""

from testtools.matchers import Equals, MatchesStructure

from .. import users
from ...testing import make_name_without_spaces, pick_bool, TestCase
from ..testing import bind


def make_origin():
    # Create a new origin with Users and User. The former refers to the
    # latter via the origin, hence why it must be bound.
    return bind(users.Users, users.User)


class TestUser(TestCase):
    def test__string_representation_includes_username_only(self):
        user = users.User(
            {
                "username": make_name_without_spaces("username"),
                "email": make_name_without_spaces("user@"),
                "is_superuser": False,
            }
        )
        self.assertThat(repr(user), Equals("<User username=%(username)r>" % user._data))

    def test__string_representation_includes_username_only_for_admin(self):
        user = users.User(
            {
                "username": make_name_without_spaces("username"),
                "email": make_name_without_spaces("user@"),
                "is_superuser": True,
            }
        )
        self.assertThat(
            repr(user), Equals("<Admin username=%(username)r>" % user._data)
        )


class TestUsers(TestCase):
    def test__whoami(self):
        username = make_name_without_spaces("username")
        email = make_name_without_spaces("user@")
        is_admin = pick_bool()

        Users = make_origin().Users
        Users._handler.whoami.return_value = {
            "username": username,
            "email": email,
            "is_superuser": is_admin,
        }

        user = Users.whoami()
        self.assertThat(
            user,
            MatchesStructure.byEquality(
                username=username, email=email, is_admin=is_admin
            ),
        )

    def test__create_without_email(self):
        username = make_name_without_spaces("username")
        password = make_name_without_spaces("password")
        email = "%s@null.maas.io" % username
        is_admin = pick_bool()

        Users = make_origin().Users
        Users._handler.create.return_value = {
            "username": username,
            "email": email,
            "is_superuser": is_admin,
        }

        Users.create(username, password, is_admin=is_admin)
        Users._handler.create.assert_called_once_with(
            username=username,
            password=password,
            email=email,
            is_superuser="1" if is_admin else "0",
        )

    def test__create_with_email(self):
        username = make_name_without_spaces("username")
        password = make_name_without_spaces("password")
        domain = "%s.com" % make_name_without_spaces("domain")
        email = "%s@%s" % (username, domain)
        is_admin = pick_bool()

        Users = make_origin().Users
        Users._handler.create.return_value = {
            "username": username,
            "email": email,
            "is_superuser": is_admin,
        }

        Users.create(username, password, email=email, is_admin=is_admin)
        Users._handler.create.assert_called_once_with(
            username=username,
            password=password,
            email=email,
            is_superuser="1" if is_admin else "0",
        )



# Copyright 2016 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MAAS CLI authentication."""

__all__ = ["obtain_credentials", "try_getpass"]

from getpass import getpass
import sys

from .creds import Credentials


def try_getpass(prompt):
    """Call `getpass`, ignoring EOF errors."""
    try:
        return getpass(prompt)
    except EOFError:
        return None


def obtain_credentials(credentials):
    """Prompt for credentials if possible.

    If the credentials are "-" then read from stdin without interactive
    prompting.
    """
    if credentials == "-":
        credentials = sys.stdin.readline().strip()
    elif credentials is None:
        credentials = try_getpass("API key (leave empty for anonymous access): ")
    # Ensure that the credentials have a valid form.
    if credentials and not credentials.isspace():
        return Credentials.parse(credentials)
    else:
        return None



"""Tests for `maas.client.viscera.domains`."""

import random

from testtools.matchers import Equals, IsInstance, MatchesStructure

from .. import domains

from ..testing import bind
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Domain and Domains. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(domains.Domains, domains.Domain)


class TestDomains(TestCase):
    def test__domains_create(self):
        origin = make_origin()
        domain_id = random.randint(0, 100)
        ttl = random.randint(0, 100)
        name = make_string_without_spaces()
        origin.Domains._handler.create.return_value = {
            "id": domain_id,
            "name": name,
            "ttl": ttl,
            "authoritative": False,
        }
        domain = origin.Domains.create(name=name, authoritative=False, ttl=ttl)
        origin.Domains._handler.create.assert_called_once_with(
            name=name, authoritative=False, ttl=ttl
        )
        self.assertThat(domain, IsInstance(origin.Domain))
        self.assertThat(
            domain,
            MatchesStructure.byEquality(
                id=domain_id, name=name, ttl=ttl, authoritative=False
            ),
        )

    def test__domains_create_without_ttl(self):
        origin = make_origin()
        domain_id = random.randint(0, 100)
        name = make_string_without_spaces()
        origin.Domains._handler.create.return_value = {"id": domain_id, "name": name}
        domain = origin.Domains.create(name=name)
        origin.Domains._handler.create.assert_called_once_with(
            name=name, authoritative=True
        )
        self.assertThat(domain, IsInstance(origin.Domain))
        self.assertThat(domain, MatchesStructure.byEquality(id=domain_id, name=name))

    def test__domains_read(self):
        """Domains.read() returns a list of Domains."""
        Domains = make_origin().Domains
        domains = [
            {"id": random.randint(0, 100), "name": make_string_without_spaces()}
            for _ in range(3)
        ]
        Domains._handler.read.return_value = domains
        domains = Domains.read()
        self.assertThat(len(domains), Equals(3))


class TestDomain(TestCase):
    def test__domain_read(self):
        Domain = make_origin().Domain
        domain = {"id": random.randint(0, 100), "name": make_string_without_spaces()}
        Domain._handler.read.return_value = domain
        self.assertThat(Domain.read(id=domain["id"]), Equals(Domain(domain)))
        Domain._handler.read.assert_called_once_with(id=domain["id"])

    def test__domain_delete(self):
        Domain = make_origin().Domain
        domain = Domain(
            {
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "description": make_string_without_spaces(),
            }
        )
        domain.delete()
        Domain._handler.delete.assert_called_once_with(id=domain.id)



"""Objects for SSH Keys."""

__all__ = ["SSHKeys", "SSHKey"]

from . import check, check_optional, Object, ObjectField, ObjectSet, ObjectType


class SSHKeysType(ObjectType):
    """Metaclass for `SSHKeys`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(cls, key: str):
        """
        Create an SSH key in MAAS with the content in `key`.

        :param key: The content of the SSH key
        :type key: `str`
        :returns: The created SSH key
        :rtype: `SSHKey`
        """
        return cls._object(await cls._handler.create(key=key))


class SSHKeys(ObjectSet, metaclass=SSHKeysType):
    """The set of SSH keys stored in MAAS."""


class SSHKeyType(ObjectType):
    async def read(cls, id: int):
        """Get an `SSHKey` by its `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class SSHKey(Object, metaclass=SSHKeyType):
    """An SSH key."""

    id = ObjectField.Checked("id", check(int), readonly=True)
    key = ObjectField.Checked("key", check(str), readonly=True)
    keysource = ObjectField.Checked("keysource", check_optional(str), readonly=True)



"""Objects for files."""

__all__ = ["File", "Files"]

from . import check, Object, ObjectField, ObjectSet, ObjectType


class FilesType(ObjectType):
    """Metaclass for `Files`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))


class Files(ObjectSet, metaclass=FilesType):
    """The set of files stored in MAAS."""


class File(Object):
    """A file stored in MAAS."""

    filename = ObjectField.Checked("filename", check(str), readonly=True)



"""Objects for zones."""

__all__ = ["Zone", "Zones"]

from . import check, Object, ObjectField, ObjectSet, ObjectType


class ZonesType(ObjectType):
    """Metaclass for `Zones`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(cls, name: str, description: str = None):
        """
        Create a `Zone` in MAAS.

        :param name: The name of the `Zone`.
        :type name: `str`
        :param description: A description of the `Zone`.
        :type description: `str`
        :returns: The create `Zone`
        :rtype: `Zone`
        """
        params = {"name": name}
        if description is not None:
            params["description"] = description
        return cls._object(await cls._handler.create(**params))


class Zones(ObjectSet, metaclass=ZonesType):
    """The set of zones stored in MAAS."""


class ZoneType(ObjectType):
    async def read(cls, name):
        data = await cls._handler.read(name=name)
        return cls(data)


class Zone(Object, metaclass=ZoneType):
    """A zone stored in MAAS."""

    id = ObjectField.Checked("id", check(int), readonly=True)

    name = ObjectField.Checked("name", check(str), check(str), pk=True)
    description = ObjectField.Checked("description", check(str), check(str))

    def __repr__(self):
        return super(Zone, self).__repr__(fields={"name", "description"})

    async def delete(self):
        """
        Deletes the `Zone` from MAAS.
        """
        return await self._handler.delete(name=self.name)



"""Commands for vlans."""

__all__ = ["register"]

from http import HTTPStatus

from . import CommandError, OriginPagedTableCommand, tables
from ..bones import CallError
from ..utils.maas_async import asynchronous


class cmd_vlans(OriginPagedTableCommand):
    """List vlans."""

    def __init__(self, parser):
        super(cmd_vlans, self).__init__(parser)
        parser.add_argument("fabric", nargs=1, help=("Name of the fabric."))
        parser.add_argument(
            "--minimal", action="store_true", help=("Output only the VIDs.")
        )

    @asynchronous
    async def load_object_sets(self, origin):
        fabrics = origin.Fabrics.read()
        subnets = origin.Subnets.read()
        return await fabrics, await subnets

    def execute(self, origin, options, target):
        visible_columns = None
        if options.minimal:
            visible_columns = ("vid",)
        try:
            fabric = origin.Fabric.read(options.fabric[0])
        except CallError as error:
            if error.status == HTTPStatus.NOT_FOUND:
                raise CommandError("Unable to find fabric %s." % options.fabric[0])
            else:
                raise
        fabrics, subnets = self.load_object_sets(origin)
        table = tables.VlansTable(
            visible_columns=visible_columns, fabrics=fabrics, subnets=subnets
        )
        return table.render(target, fabric.vlans)


class cmd_vlan(OriginPagedTableCommand):
    """Details of a vlan."""

    def __init__(self, parser):
        super(cmd_vlan, self).__init__(parser)
        parser.add_argument("fabric", nargs=1, help=("Name of the fabric."))
        parser.add_argument("vid", nargs=1, help=("VID of the VLAN."))

    @asynchronous
    async def load_object_sets(self, origin):
        fabrics = origin.Fabrics.read()
        subnets = origin.Subnets.read()
        return await fabrics, await subnets

    def get_vlan(self, vlans, vid):
        for vlan in vlans:
            if vlan.vid == vid:
                return vlan
        for vlan in vlans:
            if vlan.id == vid:
                return vlan

    def execute(self, origin, options, target):
        try:
            fabric = origin.Fabric.read(options.fabric[0])
        except CallError as error:
            if error.status == HTTPStatus.NOT_FOUND:
                raise CommandError("Unable to find fabric %s." % options.fabric[0])
            else:
                raise
        vlan_id = options.vid[0]
        if vlan_id != "untagged":
            try:
                vlan_id = int(vlan_id)
            except ValueError:
                vlan = None
            else:
                vlan = self.get_vlan(fabric.vlans, options.vid[0])
        else:
            vlan = fabric.vlans.get_default()
        if vlan is None:
            raise CommandError(
                "Unable to find VLAN %s on fabric %s."
                % (options.vid[0], options.fabric[0])
            )
        fabrics, subnets = self.load_object_sets(origin)
        table = tables.VlanDetail(fabrics=fabrics, subnets=subnets)
        return table.render(target, vlan)


def register(parser):
    """Register commands with the given parser."""
    cmd_vlans.register(parser)
    cmd_vlan.register(parser)



"""Test for `maas.client.viscera.fabrics`."""

import random

from testtools.matchers import (
    Equals,
    IsInstance,
    MatchesAll,
    MatchesSetwise,
    MatchesStructure,
)

from ...errors import CannotDelete
from ..fabrics import Fabric, Fabrics
from ..vlans import Vlan, Vlans

from ..testing import bind
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Fabrics, Fabric, Vlans, and Vlan.
    """
    return bind(Fabrics, Fabric, Vlans, Vlan)


class TestFabrics(TestCase):
    def test__fabrics_create(self):
        Fabrics = make_origin().Fabrics
        name = make_string_without_spaces()
        description = make_string_without_spaces()
        class_type = make_string_without_spaces()
        Fabrics._handler.create.return_value = {
            "id": 1,
            "name": name,
            "description": description,
            "class_type": class_type,
        }
        Fabrics.create(name=name, description=description, class_type=class_type)
        Fabrics._handler.create.assert_called_once_with(
            name=name, description=description, class_type=class_type
        )

    def test__fabrics_read(self):
        """Fabrics.read() returns a list of Fabrics."""
        Fabrics = make_origin().Fabrics
        fabrics = [
            {
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "class_type": make_string_without_spaces(),
            }
            for _ in range(3)
        ]
        Fabrics._handler.read.return_value = fabrics
        fabrics = Fabrics.read()
        self.assertThat(len(fabrics), Equals(3))


class TestFabric(TestCase):
    def test__fabric_get_default(self):
        Fabric = make_origin().Fabric
        Fabric._handler.read.return_value = {
            "id": 0,
            "name": make_string_without_spaces(),
            "class_type": make_string_without_spaces(),
        }
        Fabric.get_default()
        Fabric._handler.read.assert_called_once_with(id=0)

    def test__fabric_read(self):
        Fabric = make_origin().Fabric
        fabric = {
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "class_type": make_string_without_spaces(),
            "vlans": [{"id": 1}, {"id": 2}],
        }
        Fabric._handler.read.return_value = fabric
        self.assertThat(Fabric.read(id=fabric["id"]), Equals(Fabric(fabric)))
        Fabric._handler.read.assert_called_once_with(id=fabric["id"])
        self.assertThat(
            Fabric(fabric).vlans,
            MatchesSetwise(
                MatchesAll(IsInstance(Vlan), MatchesStructure.byEquality(id=1)),
                MatchesAll(IsInstance(Vlan), MatchesStructure.byEquality(id=2)),
            ),
        )

    def test__fabric_delete(self):
        Fabric = make_origin().Fabric
        fabric_id = random.randint(1, 100)
        fabric = Fabric(
            {
                "id": fabric_id,
                "name": make_string_without_spaces(),
                "class_type": make_string_without_spaces(),
            }
        )
        fabric.delete()
        Fabric._handler.delete.assert_called_once_with(id=fabric_id)

    def test__fabric_delete_default(self):
        Fabric = make_origin().Fabric
        fabric = Fabric(
            {
                "id": 0,
                "name": make_string_without_spaces(),
                "class_type": make_string_without_spaces(),
            }
        )
        self.assertRaises(CannotDelete, fabric.delete)



"""Testing framework for `maas.client`."""

__all__ = [
    "AsyncAwaitableMock",
    "AsyncCallableMock",
    "AsyncContextMock",
    "AsyncIterableMock",
    "make_bytes",
    "make_mac_address",
    "make_name",
    "make_name_without_spaces",
    "make_range",
    "make_string",
    "make_string_without_spaces",
    "pick_bool",
    "TestCase",
]

import asyncio
import doctest
from functools import partial
from itertools import islice, repeat
from pathlib import Path
import random
import string
from unittest import mock

from fixtures import TempDir
import testscenarios
from testtools import testcase
from testtools.matchers import DocTestMatches

from ..utils.maas_async import Asynchronous


random_letters = map(random.choice, repeat(string.ascii_letters + string.digits))

random_letters_with_spaces = map(
    random.choice, repeat(string.ascii_letters + string.digits + " ")
)

random_octet = partial(random.randint, 0, 255)

random_octets = iter(random_octet, None)


def make_string(size=10):
    """Make a random human-readable string."""
    return "".join(islice(random_letters_with_spaces, size))


def make_string_without_spaces(size=10):
    """Make a random human-readable string WITHOUT spaces."""
    return "".join(islice(random_letters, size))


def make_name(prefix="name", sep="-", size=6):
    """Make a random name.

    :param prefix: Optional prefix. Defaults to "name".
    :param sep: Separator that will go between the prefix and the random
        portion of the name. Defaults to a dash.
    :param size: Length of the random portion of the name.
    :return: A randomized unicode string.
    """
    return prefix + sep + make_string(size)


def make_name_without_spaces(prefix="name", sep="-", size=6):
    """Make a random name WITHOUT spaces.

    :param prefix: Optional prefix. Defaults to "name".
    :param sep: Separator that will go between the prefix and the random
        portion of the name. Defaults to a dash.
    :param size: Length of the random portion of the name.
    :return: A randomized unicode string.
    """
    return prefix + sep + make_string_without_spaces(size)


def make_bytes(size=10):
    """Make a random byte string."""
    return bytes(islice(random_octets, size))


def make_mac_address(delimiter=":"):
    """Make a MAC address string with the given delimiter."""
    octets = islice(random_octets, 6)
    return delimiter.join(format(octet, "02x") for octet in octets)


def make_range(cmin=1, cmax=9):
    """Return a range of random length between `cmin` and `cmax`."""
    return range(random.randint(cmin, cmax))


def pick_bool():
    """Return either `True` or `False` at random."""
    return random.choice((True, False))


class WithScenarios(testscenarios.WithScenarios):
    """Variant of testscenarios_' that provides ``__call__``."""

    def __call__(self, result=None):
        if self._get_scenarios():
            for test in testscenarios.generate_scenarios(self):
                test.__call__(result)
        else:
            super(WithScenarios, self).__call__(result)


class TestCase(WithScenarios, testcase.TestCase, metaclass=Asynchronous):
    """Base test case class for all of python-libmaas.

    This creates a new `asyncio` event loop for every test, and tears it down
    afterwards. It transparently copes with asynchronous test methods and
    other test methods by running them in this freshly created loop.
    """

    def setUp(self):
        super(TestCase, self).setUp()
        self.loop = asyncio.new_event_loop()
        self.addCleanup(self.loop.close)
        asyncio.set_event_loop(self.loop)

    def makeDir(self):
        """Create a temporary directory.

        This creates a new temporary directory. This will be removed during
        test tear-down.

        :return: The path to the directory, as a `pathlib.Path`.
        """
        tempdir = self.useFixture(TempDir())
        return Path(tempdir.path)

    def makeFile(self, name=None, contents=None, location=None):
        """Create a file, and write data to it.

        This creates a new file `name` in `location` and fills it with
        `contents`. This file will be removed during test tear-down.

        :param name: Name for the file; optional. If omitted, a random name
            will be chosen.
        :param contents: Contents for the file; optional. If omitted, some
            arbitrary bytes will be written.
        :param location: Path to a directory; optional. If omitted, a new
            temporary directory will be created with `makeDir`.

        :return: The path to the file, as a `pathlib.Path`.
        """
        location = self.makeDir() if location is None else Path(location)
        filepath = location.joinpath(make_string() if name is None else name)
        filepath.write_bytes(make_bytes() if contents is None else contents)
        return filepath

    def assertDocTestMatches(self, expected, observed, flags=None):
        """See if `observed` matches `expected`, a doctest sample.

        By default uses the doctest flags `NORMALIZE_WHITESPACE` and
        `ELLIPSIS`.
        """
        if flags is None:
            flags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
        self.assertThat(observed, DocTestMatches(expected, flags))

    def patch(self, obj, attribute, value=mock.sentinel.unset):
        """Patch `obj.attribute` with `value`.

        If `value` is unspecified, a new `Mock` will be created and patched-in
        instead. Its ``__name__`` attribute will be set to `attribute`.

        This is a thin customisation of `testtools.TestCase.patch`, so refer
        to that in case of doubt.

        :return: The patched-in object.
        """
        if value is mock.sentinel.unset:
            value = mock.Mock(__name__=attribute)
        super(TestCase, self).patch(obj, attribute, value)
        return value


class AsyncAwaitableMock(mock.Mock):
    """Mock that is "future-like"; see PEP-492.

    The new `await` syntax chokes on arguments that are not future-like, i.e.
    have an `__await__` call, so we have to fool it.

    This passes calls to `__await__` through to `__call__`.
    """

    async def __await__(_mock_self, *args, **kwargs):
        return _mock_self(*args, **kwargs)


class AsyncCallableMock(mock.Mock):
    """Mock which ensures calls are "future-like"; see PEP-492.

    As in, calls to this mock return `return_value` or `side_effect` as usual,
    but these are awaitable, or native coroutines.
    """

    async def __call__(_mock_self, *args, **kwargs):
        return super().__call__(*args, **kwargs)


class AsyncContextMock(mock.Mock):
    """Mock that acts as an async context manager; see PEP-492.

    It's not enough to mock `__aenter__` and `__aexit__` because Python
    obtains these callable attributes from the context manager's *type*. See
    https://www.python.org/dev/peps/pep-0492/#new-syntax. This is consistent
    with how non-asynchronous context managers work, but it's counterintuitive
    nonetheless.

    This returns itself from `__aenter__` and `None` from `__aexit__`.
    """

    async def __aenter__(_mock_self):
        return _mock_self

    async def __aexit__(_mock_self, *exc_info):
        return None


class AsyncIterableMock(mock.Mock):
    """Mock that can be asynchronously iterated; see PEP-492.

    This returns itself from `__aiter__` and passes through calls to
    `__anext__` to `__call__`.
    """

    def __aiter__(_mock_self):
        return _mock_self

    async def __anext__(_mock_self):
        return _mock_self()



"""Tests for `maas.client.flesh.shell`."""

import random
import sys

from testtools.matchers import Contains, Equals, Is

from .. import ArgumentParser, shell
from ...testing import make_name_without_spaces, TestCase
from .testing import capture_parse_error


class TestShell(TestCase):
    """Tests for `cmd_shell`."""

    def setUp(self):
        super(TestShell, self).setUp()
        # Start with no profiles, and no profile default.
        self.patch(shell.cmd_shell, "profile_name_choices", ())
        self.patch(shell.cmd_shell, "profile_name_default", None)

    def test_offers_profile_name_option_when_no_profiles_exist(self):
        parser = ArgumentParser()
        subparser = shell.cmd_shell.register(parser)

        # By default, the profile_name option is None.
        options = subparser.parse_args([])
        self.assertThat(options.profile_name, Is(None))

        # But any profile name can be given.
        profile_name = make_name_without_spaces("profile-name")
        options = subparser.parse_args(["--profile-name", profile_name])
        self.assertThat(options.profile_name, Equals(profile_name))

    def test_offers_profile_name_option_when_profiles_exist(self):
        profile_name_choices = tuple(
            make_name_without_spaces("profile-name") for _ in range(5)
        )
        profile_name_default = random.choice(profile_name_choices)

        self.patch(shell.cmd_shell, "profile_name_choices", profile_name_choices)
        self.patch(shell.cmd_shell, "profile_name_default", profile_name_default)

        parser = ArgumentParser()
        subparser = shell.cmd_shell.register(parser)

        # By default, the profile_name option is profile_name_default.
        options = subparser.parse_args([])
        self.assertThat(options.profile_name, Equals(profile_name_default))

        # But any profile name in profile_name_choices can be given.
        profile_name = random.choice(profile_name_choices)
        options = subparser.parse_args(["--profile-name", profile_name])
        self.assertThat(options.profile_name, Equals(profile_name))

        # Other profile names are not permitted.
        self.patch(subparser, "error").side_effect = Exception
        profile_name = make_name_without_spaces("foo")
        error = capture_parse_error(subparser, "--profile-name", profile_name)
        self.assertThat(
            str(error), Contains("argument --profile-name: invalid choice: ")
        )

    def electAttribute(self):
        # We're going to use an attribute in this module as the means for an
        # external script to report back, so choose an unlikely-to-be-in-use
        # attribute.
        module = sys.modules[__name__]
        attrname = make_name_without_spaces("attr", sep="_")
        self.patch(module, attrname, None)
        return module, attrname

    def callShell(self, *options):
        parser = ArgumentParser()
        subparser = shell.cmd_shell.register(parser)
        options = subparser.parse_args(list(options))
        options.execute(options)

    def test_runs_script_when_specified(self):
        module, attrname = self.electAttribute()

        # Mimic a non-interactive invocation of `maas shell` with a script.
        source = "import %s as mod; mod.%s = __file__" % (__name__, attrname)
        script = self.makeFile("script.py", source.encode("utf-8"))
        self.callShell(str(script))

        # That attribute has been updated.
        self.assertThat(getattr(module, attrname), Equals(str(script)))

    def test_runs_stdin_when_not_interactive(self):
        module, attrname = self.electAttribute()

        # Mimic a non-interactive invocation of `maas shell`.
        self.patch(shell, "sys")
        shell.sys.stdin.isatty.return_value = False
        shell.sys.stdin.read.return_value = "import %s as mod; mod.%s = __file__" % (
            __name__,
            attrname,
        )
        self.callShell()

        # That attribute has been updated.
        self.assertThat(getattr(module, attrname), Equals("<stdin>"))



"""Tests for `maas.client`."""

from inspect import signature
from unittest.mock import sentinel

from testtools.matchers import Equals, Is, IsInstance, Not

from .. import facade
from ... import client
from ..testing import AsyncCallableMock, TestCase
from ..viscera import Origin


class TestFunctions(TestCase):
    """Tests for module functions."""

    def test__connect_matches_Origin_connect(self):
        stub, real = client.connect, Origin.connect
        self.assertSignaturesMatch(stub, real)

    def test__connect_calls_through_to_Origin(self):
        connect = self.patch(Origin, "connect", AsyncCallableMock())
        connect.return_value = sentinel.profile, sentinel.origin
        client_object = client.connect(
            sentinel.url, apikey=sentinel.apikey, insecure=sentinel.insecure
        )
        connect.assert_called_once_with(
            sentinel.url, apikey=sentinel.apikey, insecure=sentinel.insecure
        )
        self.assertThat(client_object, IsInstance(facade.Client))
        self.assertThat(client_object._origin, Is(sentinel.origin))

    def test__login_matches_Origin_login(self):
        stub, real = client.login, Origin.login
        self.assertSignaturesMatch(stub, real)

    def test__login_calls_through_to_Origin(self):
        login = self.patch(Origin, "login", AsyncCallableMock())
        login.return_value = sentinel.profile, sentinel.origin
        client_object = client.login(
            sentinel.url,
            username=sentinel.username,
            password=sentinel.password,
            insecure=sentinel.insecure,
        )
        login.assert_called_once_with(
            sentinel.url,
            username=sentinel.username,
            password=sentinel.password,
            insecure=sentinel.insecure,
        )
        self.assertThat(client_object, IsInstance(facade.Client))
        self.assertThat(client_object._origin, Is(sentinel.origin))

    def assertSignaturesMatch(self, stub, real):
        self.assertThat(stub, Not(Is(real)))
        sig_stub = signature(client.connect)
        sig_real = signature(Origin.connect)
        self.assertThat(sig_stub, Equals(sig_real))



"""Tests for `maas.client.flesh.nodes`."""

from operator import itemgetter
import yaml

from .testing import TestCaseWithProfile
from .. import ArgumentParser, nodes, tabular
from ...enum import NodeType
from ...testing import make_name_without_spaces
from ...viscera.testing import bind
from ...viscera.nodes import Node, Nodes


def make_origin():
    """Make origin for nodes."""
    return bind(Nodes, Node)


class TestNodes(TestCaseWithProfile):
    """Tests for `cmd_nodes`."""

    def test_returns_table_with_nodes(self):
        origin = make_origin()
        parser = ArgumentParser()
        node_obj = [
            {
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.MACHINE.value,
            },
            {
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.DEVICE.value,
            },
            {
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.RACK_CONTROLLER.value,
            },
            {
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.REGION_CONTROLLER.value,
            },
            {
                "hostname": make_name_without_spaces(),
                "node_type": NodeType.REGION_AND_RACK_CONTROLLER.value,
            },
        ]
        origin.Nodes._handler.read.return_value = node_obj
        cmd = nodes.cmd_nodes(parser)
        subparser = nodes.cmd_nodes.register(parser)
        options = subparser.parse_args([])
        output = yaml.safe_load(
            cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        )
        self.assertEquals(
            [
                {"name": "hostname", "title": "Hostname"},
                {"name": "node_type", "title": "Type"},
            ],
            output["columns"],
        )
        nodes_output = sorted(
            [
                {"hostname": node["hostname"], "node_type": node["node_type"]}
                for node in node_obj
            ],
            key=itemgetter("hostname"),
        )
        self.assertEquals(nodes_output, output["data"])

    def test_calls_handler_with_hostnames(self):
        origin = make_origin()
        parser = ArgumentParser()
        origin.Nodes._handler.read.return_value = []
        subparser = nodes.cmd_nodes.register(parser)
        cmd = nodes.cmd_nodes(parser)
        hostnames = [make_name_without_spaces() for _ in range(3)]
        options = subparser.parse_args(hostnames)
        cmd.execute(origin, options, target=tabular.RenderTarget.yaml)
        origin.Nodes._handler.read.assert_called_once_with(hostname=hostnames)



"""Tests for `maas.client.flesh.profiles`."""

from argparse import ArgumentParser
from io import StringIO
import sys
from textwrap import dedent
from unittest.mock import call

from .. import profiles
from ...bones.helpers import MacaroonLoginNotSupported
from ...testing import AsyncCallableMock, TestCase
from ...utils.tests.test_profiles import make_profile


class TestLogin(TestCase):
    """Tests for `cmd_login`."""

    def test_login_no_macaroons_prompts_user_pass(self):
        profile = make_profile()

        stdout = self.patch(sys, "stdout", StringIO())
        mock_read_input = self.patch(profiles, "read_input")
        mock_read_input.side_effect = ["username", "password"]
        mock_login = AsyncCallableMock(side_effect=[MacaroonLoginNotSupported, profile])
        self.patch(profiles.helpers, "login", mock_login)

        parser = ArgumentParser()
        cmd = profiles.cmd_login(parser)
        options = parser.parse_args(["http://maas.example"])
        cmd(options)
        mock_login.assert_has_calls(
            [
                call(
                    "http://maas.example/api/2.0/",
                    anonymous=False,
                    insecure=False,
                    username=None,
                    password=None,
                ),
                call(
                    "http://maas.example/api/2.0/",
                    insecure=False,
                    username="username",
                    password="password",
                ),
            ]
        )
        self.assertIn("Congratulations!", stdout.getvalue())

    def test_print_whats_next(self):
        profile = make_profile()
        stdout = self.patch(sys, "stdout", StringIO())
        profiles.cmd_login.print_whats_next(profile)
        expected = dedent(
            """\
            Congratulations! You are logged in to the MAAS
            server at {profile.url} with the profile name
            {profile.name}.

            For help with the available commands, try:

              maas help

            """
        ).format(profile=profile)
        observed = stdout.getvalue()
        self.assertDocTestMatches(expected, observed)



# Copyright 2016-2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for `maas.client.utils`."""

import base64
from functools import partial
from itertools import cycle
import os
import os.path
from unittest.mock import sentinel

from testtools.matchers import (
    AfterPreprocessing,
    Equals,
    Is,
    MatchesListwise,
    MatchesStructure,
)
from twisted.internet.task import Clock

from ... import utils
from ...testing import make_name_without_spaces, make_string, TestCase


class TestMAASOAuth(TestCase):
    def test_OAuthSigner_sign_request_adds_header(self):
        token_key = make_name_without_spaces("token-key")
        token_secret = make_name_without_spaces("token-secret")
        consumer_key = make_name_without_spaces("consumer-key")
        consumer_secret = make_name_without_spaces("consumer-secret")
        realm = make_name_without_spaces("realm")

        headers = {}
        auth = utils.OAuthSigner(
            token_key=token_key,
            token_secret=token_secret,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            realm=realm,
        )
        auth.sign_request("http://example.com/", "GET", None, headers)

        self.assertIn("Authorization", headers)
        authorization = headers["Authorization"]
        self.assertIn('realm="%s"' % realm, authorization)
        self.assertIn('oauth_token="%s"' % token_key, authorization)
        self.assertIn('oauth_consumer_key="%s"' % consumer_key, authorization)
        self.assertIn(
            'oauth_signature="%s%%26%s"' % (consumer_secret, token_secret),
            authorization,
        )

    def test_sign_adds_header(self):
        token_key = make_name_without_spaces("token-key")
        token_secret = make_name_without_spaces("token-secret")
        consumer_key = make_name_without_spaces("consumer-key")

        headers = {}
        utils.sign(
            "http://example.com/", headers, (consumer_key, token_key, token_secret)
        )

        self.assertIn("Authorization", headers)
        authorization = headers["Authorization"]
        self.assertIn('realm="OAuth"', authorization)
        self.assertIn('oauth_token="%s"' % token_key, authorization)
        self.assertIn('oauth_consumer_key="%s"' % consumer_key, authorization)
        self.assertIn('oauth_signature="%%26%s"' % token_secret, authorization)


class TestPayloadPreparation(TestCase):
    """Tests for `prepare_payload`."""

    uri_base = "http://example.com/MAAS/api/2.0/"

    # Scenarios for ReSTful operations; i.e. without an "op" parameter.
    scenarios_without_op = (
        # Without data, all requests have an empty request body and no extra
        # headers.
        (
            "create",
            {
                "method": "POST",
                "data": [],
                "expected_uri": uri_base,
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "read",
            {
                "method": "GET",
                "data": [],
                "expected_uri": uri_base,
                "expected_body": None,
                "expected_headers": [],
            },
        ),
        (
            "update",
            {
                "method": "PUT",
                "data": [],
                "expected_uri": uri_base,
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "delete",
            {
                "method": "DELETE",
                "data": [],
                "expected_uri": uri_base,
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        # With data, PUT, POST, and DELETE requests have their body and
        # extra headers prepared by build_multipart_message and
        # encode_multipart_message. For GET requests, the data is
        # encoded into the query string, and both the request body and
        # extra headers are empty.
        (
            "create-with-data",
            {
                "method": "POST",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base,
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "read-with-data",
            {
                "method": "GET",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base + "?foo=bar&foo=baz",
                "expected_body": None,
                "expected_headers": [],
            },
        ),
        (
            "update-with-data",
            {
                "method": "PUT",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base,
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "delete-with-data",
            {
                "method": "DELETE",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base,
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
    )

    # Scenarios for non-ReSTful operations; i.e. with an "op" parameter.
    scenarios_with_op = (
        # Without data, all requests have an empty request body and no extra
        # headers. The operation is encoded into the query string.
        (
            "create",
            {
                "method": "POST",
                "data": [],
                "expected_uri": uri_base + "?op=something",
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "read",
            {
                "method": "GET",
                "data": [],
                "expected_uri": uri_base + "?op=something",
                "expected_body": None,
                "expected_headers": [],
            },
        ),
        (
            "update",
            {
                "method": "PUT",
                "data": [],
                "expected_uri": uri_base + "?op=something",
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "delete",
            {
                "method": "DELETE",
                "data": [],
                "expected_uri": uri_base + "?op=something",
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        # With data, PUT, POST, and DELETE requests have their body and
        # extra headers prepared by build_multipart_message and
        # encode_multipart_message. For GET requests, the data is
        # encoded into the query string, and both the request body and
        # extra headers are empty. The operation is encoded into the
        # query string.
        (
            "create-with-data",
            {
                "method": "POST",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base + "?op=something",
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "read-with-data",
            {
                "method": "GET",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base + "?op=something&foo=bar&foo=baz",
                "expected_body": None,
                "expected_headers": [],
            },
        ),
        (
            "update-with-data",
            {
                "method": "PUT",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base + "?op=something",
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
        (
            "delete-with-data",
            {
                "method": "DELETE",
                "data": [("foo", "bar"), ("foo", "baz")],
                "expected_uri": uri_base + "?op=something",
                "expected_body": sentinel.body,
                "expected_headers": sentinel.headers,
            },
        ),
    )

    scenarios_without_op = tuple(
        ("%s-without-op" % name, dict(scenario, op=None))
        for name, scenario in scenarios_without_op
    )

    scenarios_with_op = tuple(
        ("%s-with-op" % name, dict(scenario, op="something"))
        for name, scenario in scenarios_with_op
    )

    scenarios = scenarios_without_op + scenarios_with_op

    def test_prepare_payload(self):
        # Patch build_multipart_message and encode_multipart_message to
        # match the scenarios.
        build_multipart = self.patch(utils, "build_multipart_message")
        build_multipart.return_value = sentinel.message
        encode_multipart = self.patch(utils, "encode_multipart_message")
        encode_multipart.return_value = sentinel.headers, sentinel.body
        # The payload returned is a 3-tuple of (uri, body, headers). Pass
        # `data` as an iterator to ensure that it works with non-sized types.
        payload = utils.prepare_payload(
            op=self.op, method=self.method, uri=self.uri_base, data=iter(self.data)
        )
        expected = (
            Equals(self.expected_uri),
            Equals(self.expected_body),
            Equals(self.expected_headers),
        )
        self.assertThat(payload, MatchesListwise(expected))
        # encode_multipart_message, when called, is passed the data
        # unadulterated.
        if self.expected_body is sentinel.body:
            encode_multipart.assert_called_once_with(sentinel.message)


class TestPayloadPreparationWithFiles(TestCase):
    """Tests for `prepare_payload` involving files."""

    def test_files_are_included(self):
        parameter = make_string()
        contents = os.urandom(5)
        filepath = self.makeFile(contents=contents)
        # Writing the parameter as "parameter@=filename" on the
        # command-line causes name_value_pair() to return a `name,
        # opener` tuple, where `opener` is a callable that returns an
        # open file handle.
        data = [(parameter, partial(filepath.open, "rb"))]
        uri, body, headers = utils.prepare_payload(
            op=None, method="POST", uri="http://localhost", data=data
        )

        expected_body_template = """\
            --...
            Content-Transfer-Encoding: base64
            Content-Disposition: form-data; ...name="%s"; ...name="%s"
            MIME-Version: 1.0
            Content-Type: application/octet-stream

            %s
            --...--
            """
        expected_body = expected_body_template % (
            parameter,
            parameter,
            base64.b64encode(contents).decode("ascii"),
        )

        self.assertDocTestMatches(expected_body, body.decode("ascii"))


class TestDocstringParsing(TestCase):
    """Tests for docstring parsing with `parse_docstring`."""

    def test_basic(self):
        self.assertEqual(("Title", "Body"), utils.parse_docstring("Title\n\nBody"))
        self.assertEqual(
            ("A longer title", "A longer body"),
            utils.parse_docstring("A longer title\n\nA longer body"),
        )

    def test_returns_named_tuple(self):
        self.assertThat(
            utils.parse_docstring("Title\n\nBody"),
            MatchesStructure.byEquality(title="Title", body="Body"),
        )

    def test_no_body(self):
        # parse_docstring returns an empty string when there's no body.
        self.assertEqual(("Title", ""), utils.parse_docstring("Title\n\n"))
        self.assertEqual(("Title", ""), utils.parse_docstring("Title"))

    def test_unwrapping(self):
        # parse_docstring unwraps the title paragraph, and dedents the body
        # paragraphs.
        self.assertEqual(
            (
                "Title over two lines",
                "Paragraph over\ntwo lines\n\n" "Another paragraph\nover two lines",
            ),
            utils.parse_docstring(
                """
                Title over
                two lines

                Paragraph over
                two lines

                Another paragraph
                over two lines
                """
            ),
        )

    def test_gets_docstring_from_function(self):
        # parse_docstring can extract the docstring when the argument passed
        # is not a string type.
        def example():
            """Title.

            Body.
            """

        self.assertEqual(("Title.", "Body."), utils.parse_docstring(example))

    def test_normalises_whitespace(self):
        # parse_docstring can parse CRLF/CR/LF text, but always emits LF (\n,
        # new-line) separated text.
        self.assertEqual(("long title", ""), utils.parse_docstring("long\r\ntitle"))
        self.assertEqual(
            ("title", "body1\n\nbody2"),
            utils.parse_docstring("title\n\nbody1\r\rbody2"),
        )


class TestFunctions(TestCase):
    """Tests for miscellaneous functions in `maas.client.utils`."""

    def test_api_url(self):
        transformations = list(
            {
                "http://example.com/": "http://example.com/api/2.0/",
                "http://example.com/foo": "http://example.com/foo/api/2.0/",
                "http://example.com/foo/": "http://example.com/foo/api/2.0/",
                "http://example.com/api/7.9": "http://example.com/api/7.9/",
                "http://example.com/api/7.9/": "http://example.com/api/7.9/",
            }.items()
        )
        urls = [url for url, url_out in transformations]
        urls_out = [url_out for url, url_out in transformations]
        expected = [
            AfterPreprocessing(utils.api_url, Equals(url_out)) for url_out in urls_out
        ]
        self.assertThat(urls, MatchesListwise(expected))

    def test_coalesce(self):
        self.assertThat(utils.coalesce("abc"), Equals("abc"))
        self.assertThat(utils.coalesce(None, "abc"), Equals("abc"))
        self.assertThat(utils.coalesce("abc", None), Equals("abc"))
        self.assertThat(utils.coalesce("abc", "def"), Equals("abc"))
        self.assertThat(utils.coalesce(default="foo"), Equals("foo"))
        self.assertThat(utils.coalesce(None, default="foo"), Equals("foo"))
        self.assertThat(utils.coalesce(), Is(None))


class TestRetries(TestCase):
    def assertRetry(
        self, clock, observed, expected_elapsed, expected_remaining, expected_wait
    ):
        """Assert that the retry tuple matches the given expectations.

        Retry tuples are those returned by `retries`.
        """
        self.assertThat(
            observed,
            MatchesListwise(
                [
                    Equals(expected_elapsed),  # elapsed
                    Equals(expected_remaining),  # remaining
                    Equals(expected_wait),  # wait
                ]
            ),
        )

    def test_yields_elapsed_remaining_and_wait(self):
        # Take control of time.
        clock = Clock()

        gen_retries = utils.retries(5, 2, time=clock.seconds)
        # No time has passed, 5 seconds remain, and it suggests sleeping
        # for 2 seconds.
        self.assertRetry(clock, next(gen_retries), 0, 5, 2)
        # Mimic sleeping for the suggested sleep time.
        clock.advance(2)
        # Now 2 seconds have passed, 3 seconds remain, and it suggests
        # sleeping for 2 more seconds.
        self.assertRetry(clock, next(gen_retries), 2, 3, 2)
        # Mimic sleeping for the suggested sleep time.
        clock.advance(2)
        # Now 4 seconds have passed, 1 second remains, and it suggests
        # sleeping for just 1 more second.
        self.assertRetry(clock, next(gen_retries), 4, 1, 1)
        # Mimic sleeping for the suggested sleep time.
        clock.advance(1)
        # There's always a final chance to try something.
        self.assertRetry(clock, next(gen_retries), 5, 0, 0)
        # All done.
        self.assertRaises(StopIteration, next, gen_retries)

    def test_calculates_times_with_reference_to_current_time(self):
        # Take control of time.
        clock = Clock()

        gen_retries = utils.retries(5, 2, time=clock.seconds)
        # No time has passed, 5 seconds remain, and it suggests sleeping
        # for 2 seconds.
        self.assertRetry(clock, next(gen_retries), 0, 5, 2)
        # Mimic sleeping for 4 seconds, more than the suggested.
        clock.advance(4)
        # Now 4 seconds have passed, 1 second remains, and it suggests
        # sleeping for just 1 more second.
        self.assertRetry(clock, next(gen_retries), 4, 1, 1)
        # Don't sleep, ask again immediately, and the same answer is given.
        self.assertRetry(clock, next(gen_retries), 4, 1, 1)
        # Mimic sleeping for 100 seconds, much more than the suggested.
        clock.advance(100)
        # There's always a final chance to try something, but the elapsed and
        # remaining figures are still calculated with reference to the current
        # time. The wait time never goes below zero.
        self.assertRetry(clock, next(gen_retries), 104, -99, 0)
        # All done.
        self.assertRaises(StopIteration, next, gen_retries)

    def test_captures_start_time_when_called(self):
        # Take control of time.
        clock = Clock()

        gen_retries = utils.retries(5, 2, time=clock.seconds)
        clock.advance(4)
        # 4 seconds have passed, so 1 second remains, and it suggests sleeping
        # for 1 second.
        self.assertRetry(clock, next(gen_retries), 4, 1, 1)

    def test_intervals_can_be_an_iterable(self):
        # Take control of time.
        clock = Clock()
        # Use intervals of 1s, 2s, 3, and then back to 1s.
        intervals = cycle((1.0, 2.0, 3.0))

        gen_retries = utils.retries(5, intervals, time=clock.seconds)
        # No time has passed, 5 seconds remain, and it suggests sleeping
        # for 1 second, then 2, then 3, then 1 again.
        self.assertRetry(clock, next(gen_retries), 0, 5, 1)
        self.assertRetry(clock, next(gen_retries), 0, 5, 2)
        self.assertRetry(clock, next(gen_retries), 0, 5, 3)
        self.assertRetry(clock, next(gen_retries), 0, 5, 1)
        # Mimic sleeping for 3.5 seconds, more than the suggested.
        clock.advance(3.5)
        # Now 3.5 seconds have passed, 1.5 seconds remain, and it suggests
        # sleeping for 1.5 seconds, 0.5 less than the next expected interval
        # of 2.0 seconds.
        self.assertRetry(clock, next(gen_retries), 3.5, 1.5, 1.5)
        # Don't sleep, ask again immediately, and the same answer is given.
        self.assertRetry(clock, next(gen_retries), 3.5, 1.5, 1.5)
        # Don't sleep, ask again immediately, and 1.0 seconds is given,
        # because we're back to the 1.0 second interval.
        self.assertRetry(clock, next(gen_retries), 3.5, 1.5, 1.0)
        # Mimic sleeping for 100 seconds, much more than the suggested.
        clock.advance(100)
        # There's always a final chance to try something, but the elapsed and
        # remaining figures are still calculated with reference to the current
        # time. The wait time never goes below zero.
        self.assertRetry(clock, next(gen_retries), 103.5, -98.5, 0.0)
        # All done.
        self.assertRaises(StopIteration, next, gen_retries)


class TestRemoveNone(TestCase):
    """Test `remove_None`."""

    def test_removes_all_None_values(self):
        data = {"None": None, "another_None": None, "keep": "value"}
        self.assertEquals({"keep": "value"}, utils.remove_None(data))



# Copyright 2017 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for handling the calculation of object difference."""

import copy

from ...testing import TestCase
from ..diff import calculate_dict_diff


class TestCalculateDictDiff(TestCase):
    """Test `calculate_dict_diff`."""

    def test_calcs_no_difference(self):
        orig_data = {"key1": "value1", "key2": "value2"}
        new_data = copy.deepcopy(orig_data)
        self.assertEquals({}, calculate_dict_diff(orig_data, new_data))

    def test_calcs_changed_value(self):
        orig_data = {"key1": "value1", "key2": "value2"}
        new_data = copy.deepcopy(orig_data)
        new_data["key2"] = "new_value"
        self.assertEquals(
            {"key2": "new_value"}, calculate_dict_diff(orig_data, new_data)
        )

    def test_calcs_deleted_value(self):
        orig_data = {"key1": "value1", "key2": "value2"}
        new_data = copy.deepcopy(orig_data)
        del new_data["key2"]
        self.assertEquals({"key2": ""}, calculate_dict_diff(orig_data, new_data))

    def test_calcs_changes_and_deleted(self):
        orig_data = {"key1": "value1", "key2": "value2"}
        new_data = copy.deepcopy(orig_data)
        new_data["key1"] = "new_value"
        del new_data["key2"]
        self.assertEquals(
            {"key1": "new_value", "key2": ""}, calculate_dict_diff(orig_data, new_data)
        )



"""Tests for `maas.client.bones.helpers`."""

import json
from unittest.mock import Mock
from urllib.parse import urlparse, urlsplit

import aiohttp.web
from macaroonbakery.httpbakery import Client
from testtools import ExpectedException
from testtools.matchers import Equals, Is, IsInstance, MatchesDict

from .. import helpers, testing
from ...testing import AsyncCallableMock, make_name, make_name_without_spaces, TestCase
from ...utils import api_url, profiles
from ...utils.testing import make_Credentials
from ..testing import api_descriptions
from ..testing.server import ApplicationBuilder


class TestFetchAPIDescription(TestCase):
    """Tests for `fetch_api_description`."""

    def test__raises_RemoteError_when_request_fails(self):
        fixture = self.useFixture(testing.DescriptionServer(b"bogus"))
        error = self.assertRaises(
            helpers.RemoteError,
            self.loop.run_until_complete,
            helpers.fetch_api_description(fixture.url + "bogus/"),
        )
        self.assertEqual(fixture.url + "bogus/ -> 404 Not Found", str(error))

    def test__raises_RemoteError_when_content_not_json(self):
        fixture = self.useFixture(testing.DescriptionServer())
        fixture.handler.content_type = "text/json"
        error = self.assertRaises(
            helpers.RemoteError,
            self.loop.run_until_complete,
            helpers.fetch_api_description(fixture.url),
        )
        self.assertEqual("Expected application/json, got: text/json", str(error))


class TestFetchAPIDescriptionURLs(TestCase):
    """Tests for URL types accepted by `fetch_api_description`."""

    scenarios = (
        ("string", dict(prepare=str)),
        ("split", dict(prepare=urlsplit)),
        ("parsed", dict(prepare=urlparse)),
    )

    def test__accepts_prepared_url(self):
        description = {"foo": make_name_without_spaces("bar")}
        description_json = json.dumps(description).encode("ascii")
        fixture = self.useFixture(testing.DescriptionServer(description_json))
        description_url = self.prepare(fixture.url)  # Parse, perhaps.
        description_fetched = self.loop.run_until_complete(
            helpers.fetch_api_description(description_url)
        )
        self.assertThat(description_fetched, Equals(description))


class TestFetchAPIDescription_APIVersions(TestCase):
    """Tests for `fetch_api_description` with multiple API versions."""

    scenarios = tuple(
        (name, dict(version=version, path=path))
        for name, version, path in testing.list_api_descriptions()
    )

    def test__downloads_description(self):
        description = self.path.read_bytes()
        fixture = self.useFixture(testing.DescriptionServer(description))
        description_fetched = self.loop.run_until_complete(
            helpers.fetch_api_description(fixture.url)
        )
        self.assertThat(
            description_fetched, Equals(json.loads(description.decode("utf-8")))
        )


class TestConnect(TestCase):
    """Tests for `maas.client.utils.connect.connect`."""

    def setUp(self):
        super(TestConnect, self).setUp()
        self.patch(helpers, "fetch_api_description", AsyncCallableMock(return_value={}))

    def test__anonymous(self):
        # Connect without an apikey.
        profile = helpers.connect("http://example.org:5240/MAAS/")
        helpers.fetch_api_description.assert_called_once_with(
            urlparse("http://example.org:5240/MAAS/api/2.0/"), False
        )
        # A Profile instance was returned with no credentials.
        self.assertThat(profile, IsInstance(profiles.Profile))
        self.assertThat(profile.credentials, Is(None))

    def test__connected_when_apikey_provided(self):
        credentials = make_Credentials()
        # Connect with an apikey.
        profile = helpers.connect(
            "http://example.org:5240/MAAS/", apikey=str(credentials)
        )
        # The description was fetched.
        helpers.fetch_api_description.assert_called_once_with(
            urlparse("http://example.org:5240/MAAS/api/2.0/"), False
        )
        # A Profile instance was returned with the expected credentials.
        self.assertThat(profile, IsInstance(profiles.Profile))
        self.assertThat(profile.credentials, Equals(credentials))

    def test__complains_when_username_in_URL(self):
        self.assertRaises(
            helpers.ConnectError,
            helpers.connect,
            "http://foo:bar@example.org:5240/MAAS/",
        )

    def test__complains_when_password_in_URL(self):
        self.assertRaises(
            helpers.ConnectError, helpers.connect, "http://:bar@example.org:5240/MAAS/"
        )

    def test__URL_is_normalised_to_point_at_API_endpoint(self):
        profile = helpers.connect("http://example.org:5240/MAAS/")
        self.assertThat(profile.url, Equals(api_url("http://example.org:5240/MAAS/")))

    def test__profile_is_given_default_name_based_on_URL(self):
        domain = make_name_without_spaces("domain")
        profile = helpers.connect("http://%s/MAAS/" % domain)
        self.assertThat(profile.name, Equals(domain))

    def test__API_description_is_saved_in_profile(self):
        description = helpers.fetch_api_description.return_value = {"foo": "bar"}
        profile = helpers.connect("http://example.org:5240/MAAS/")
        self.assertThat(profile.description, Equals(description))

    def test__API_description_is_fetched_insecurely_if_requested(self):
        profile = helpers.connect("http://example.org:5240/MAAS/", insecure=True)
        helpers.fetch_api_description.assert_called_once_with(
            urlparse("http://example.org:5240/MAAS/api/2.0/"), True
        )
        self.assertTrue(profile.other["insecure"])


class TestLogin(TestCase):
    """Tests for `maas.client.utils.login.login`."""

    def setUp(self):
        super(TestLogin, self).setUp()
        self.patch(helpers, "authenticate", AsyncCallableMock(return_value=None))
        self.patch(helpers, "fetch_api_description", AsyncCallableMock(return_value={}))

    def test__anonymous(self):
        # Log-in anonymously.
        profile = helpers.login("http://example.org:5240/MAAS/", anonymous=True)
        # No token was obtained, but the description was fetched.
        helpers.authenticate.assert_not_called()
        # A Profile instance was returned with no credentials.
        self.assertThat(profile, IsInstance(profiles.Profile))
        self.assertThat(profile.credentials, Is(None))

    def test__macaroon_auth_with_no_username_and_password(self):
        credentials = make_Credentials()
        self.patch(
            helpers,
            "authenticate_with_macaroon",
            AsyncCallableMock(return_value=credentials),
        )
        # Log-in without a user-name or a password.
        profile = helpers.login("http://example.org:5240/MAAS/")
        # A token is obtained via macaroons, but the description was fetched.
        # The description was fetched.
        helpers.fetch_api_description.assert_called_once_with(
            urlparse("http://example.org:5240/MAAS/api/2.0/"), False
        )
        # The returned profile uses credentials obtained from the
        # authentication
        self.assertThat(profile, IsInstance(profiles.Profile))
        self.assertThat(profile.credentials, Is(credentials))

    def test__authenticated_when_username_and_password_provided(self):
        credentials = make_Credentials()
        helpers.authenticate.return_value = credentials
        # Log-in with a user-name and a password.
        profile = helpers.login("http://foo:bar@example.org:5240/MAAS/")
        # A token was obtained, and the description was fetched.
        helpers.authenticate.assert_called_once_with(
            "http://example.org:5240/MAAS/api/2.0/", "foo", "bar", insecure=False
        )
        # A Profile instance was returned with the expected credentials.
        self.assertThat(profile, IsInstance(profiles.Profile))
        self.assertThat(profile.credentials, Is(credentials))

    def test__complains_when_username_but_not_password(self):
        self.assertRaises(
            helpers.UsernameWithoutPassword,
            helpers.login,
            "http://example.org:5240/MAAS/",
            username="alice",
        )

    def test__complains_when_password_but_not_username(self):
        self.assertRaises(
            helpers.PasswordWithoutUsername,
            helpers.login,
            "http://example.org:5240/MAAS/",
            password="wonderland",
        )

    def test__complains_when_username_in_URL_and_passed_explicitly(self):
        self.assertRaises(
            helpers.LoginError,
            helpers.login,
            "http://foo:bar@example.org:5240/MAAS/",
            username="alice",
        )

    def test__complains_when_empty_username_in_URL_and_passed_explicitly(self):
        self.assertRaises(
            helpers.LoginError,
            helpers.login,
            "http://:bar@example.org:5240/MAAS/",
            username="alice",
        )

    def test__complains_when_password_in_URL_and_passed_explicitly(self):
        self.assertRaises(
            helpers.LoginError,
            helpers.login,
            "http://foo:bar@example.org:5240/MAAS/",
            password="wonderland",
        )

    def test__complains_when_empty_password_in_URL_and_passed_explicitly(self):
        self.assertRaises(
            helpers.LoginError,
            helpers.login,
            "http://foo:@example.org:5240/MAAS/",
            password="wonderland",
        )

    def test__URL_is_normalised_to_point_at_API_endpoint(self):
        profile = helpers.login("http://example.org:5240/MAAS/", anonymous=True)
        self.assertThat(profile.url, Equals(api_url("http://example.org:5240/MAAS/")))

    def test__profile_is_given_default_name_based_on_URL(self):
        domain = make_name_without_spaces("domain")
        profile = helpers.login("http://%s/MAAS/" % domain, anonymous=True)
        self.assertThat(profile.name, Equals(domain))

    def test__API_description_is_saved_in_profile(self):
        description = {make_name("key"): make_name("value")}
        helpers.fetch_api_description.return_value = description
        profile = helpers.login("http://example.org:5240/MAAS/", anonymous=True)
        self.assertThat(profile.description, Equals(description))

    def test__API_token_is_fetched_insecurely_if_requested(self):
        profile = helpers.login("http://foo:bar@example.org:5240/MAAS/", insecure=True)
        helpers.authenticate.assert_called_once_with(
            "http://example.org:5240/MAAS/api/2.0/", "foo", "bar", insecure=True
        )
        self.assertTrue(profile.other["insecure"])

    def test__API_description_is_fetched_insecurely_if_requested(self):
        helpers.login("http://example.org:5240/MAAS/", anonymous=True, insecure=True)
        helpers.fetch_api_description.assert_called_once_with(
            urlparse("http://example.org:5240/MAAS/api/2.0/"), True
        )

    def test__uses_username_from_URL_if_set(self):
        helpers.login("http://foo@maas.io/", password="bar")
        helpers.authenticate.assert_called_once_with(
            "http://maas.io/api/2.0/", "foo", "bar", insecure=False
        )

    def test__uses_username_and_password_from_URL_if_set(self):
        helpers.login("http://foo:bar@maas.io/")
        helpers.authenticate.assert_called_once_with(
            "http://maas.io/api/2.0/", "foo", "bar", insecure=False
        )

    def test__uses_empty_username_and_password_in_URL_if_set(self):
        helpers.login("http://:@maas.io/")
        helpers.authenticate.assert_called_once_with(
            "http://maas.io/api/2.0/", "", "", insecure=False
        )


class TestAuthenticate(TestCase):
    """Tests for `authenticate`."""

    scenarios = tuple(
        (name, dict(version=version, description=description))
        for name, version, description in api_descriptions
    )

    async def test__obtains_credentials_from_server(self):
        builder = ApplicationBuilder(self.description)

        @builder.handle("anon:Version.read")
        async def version(request):
            return {"capabilities": ["authenticate-api"]}

        credentials = make_Credentials()
        parameters = None

        @builder.route("POST", "/accounts/authenticate/")
        async def deploy(request):
            nonlocal parameters
            parameters = await request.post()
            return {
                "consumer_key": credentials.consumer_key,
                "token_key": credentials.token_key,
                "token_secret": credentials.token_secret,
            }

        username = make_name_without_spaces("username")
        password = make_name_without_spaces("password")

        async with builder.serve() as baseurl:
            credentials_observed = await helpers.authenticate(
                baseurl, username, password
            )

        self.assertThat(credentials_observed, Equals(credentials))
        self.assertThat(
            parameters,
            MatchesDict(
                {
                    "username": Equals(username),
                    "password": Equals(password),
                    "consumer": IsInstance(str),
                }
            ),
        )

    async def test__raises_error_when_server_does_not_support_authn(self):
        builder = ApplicationBuilder(self.description)

        @builder.handle("anon:Version.read")
        async def version(request):
            return {"capabilities": []}

        async with builder.serve() as baseurl:
            with ExpectedException(helpers.LoginNotSupported):
                await helpers.authenticate(baseurl, "username", "password")

    async def test__raises_error_when_server_rejects_credentials(self):
        builder = ApplicationBuilder(self.description)

        @builder.handle("anon:Version.read")
        async def version(request):
            return {"capabilities": ["authenticate-api"]}

        @builder.route("POST", "/accounts/authenticate/")
        async def deploy(request):
            raise aiohttp.web.HTTPForbidden()

        async with builder.serve() as baseurl:
            with ExpectedException(helpers.RemoteError):
                await helpers.authenticate(baseurl, "username", "password")


class TestAuthenticateWithMacaroon(TestCase):
    def setUp(self):
        super().setUp()
        self.mock_client_request = self.patch(Client, "request")
        self.token_result = {
            "consumer_key": "abc",
            "token_key": "123",
            "token_secret": "xyz",
        }
        self.mock_response = Mock()
        self.mock_response.status_code = 200
        self.mock_response.json.return_value = self.token_result
        self.mock_client_request.return_value = self.mock_response

    async def test__authenticate_with_bakery_creates_token(self):
        credentials = await helpers.authenticate_with_macaroon("http://example.com")
        self.assertEqual(credentials, "abc:123:xyz")
        # a call to create an API token is made
        self.mock_client_request.assert_called_once_with(
            "POST",
            "http://example.com/account/?op=create_authorisation_token",
            verify=True,
        )

    async def test__authenticate_failed_request(self):
        self.mock_response.status_code = 500
        self.mock_response.text = "error!"
        try:
            await helpers.authenticate_with_macaroon("http://example.com")
        except helpers.LoginError as e:
            self.assertEqual(str(e), "Login failed: error!")
        else:
            self.fail("LoginError not raised")

    async def test__authenticate_macaroon_not_supported(self):
        self.mock_response.status_code = 401
        try:
            await helpers.authenticate_with_macaroon("http://example.com")
        except helpers.MacaroonLoginNotSupported as e:
            self.assertEqual(str(e), "Macaroon authentication not supported")
        else:
            self.fail("MacaroonLoginNotSupported not raised")


class TestDeriveResourceName(TestCase):
    """Tests for `derive_resource_name`."""

    def test__removes_Anon_prefix(self):
        self.assertThat(helpers.derive_resource_name("AnonFooBar"), Equals("FooBar"))

    def test__removes_Handler_suffix(self):
        self.assertThat(helpers.derive_resource_name("FooBarHandler"), Equals("FooBar"))

    def test__normalises_Maas_to_MAAS(self):
        self.assertThat(helpers.derive_resource_name("Maas"), Equals("MAAS"))

    def test__does_all_the_above(self):
        self.assertThat(helpers.derive_resource_name("AnonMaasHandler"), Equals("MAAS"))



"""Objects for block devices."""

__all__ = ["BlockDevice", "BlockDevices"]

from typing import Iterable, Union

from . import (
    check,
    check_optional,
    Object,
    ObjectField,
    ObjectFieldRelated,
    ObjectFieldRelatedSet,
    ObjectSet,
    ObjectType,
    to,
)
from .nodes import Node
from ..enum import BlockDeviceType, PartitionTableType
from ..utils import remove_None


class BlockDeviceTypeMeta(ObjectType):
    """Metaclass for `BlockDevice`."""

    async def read(cls, node, id):
        """Get `BlockDevice` by `id`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        return cls(await cls._handler.read(system_id=system_id, id=id))


class BlockDevice(Object, metaclass=BlockDeviceTypeMeta):
    """A block device on a machine."""

    node = ObjectFieldRelated("system_id", "Node", readonly=True, pk=0)
    id = ObjectField.Checked("id", check(int), readonly=True, pk=1)
    type = ObjectField.Checked("type", to(BlockDeviceType), readonly=True)
    name = ObjectField.Checked("name", check(str), check(str), alt_pk=1)
    model = ObjectField.Checked("model", check_optional(str), check_optional(str))
    serial = ObjectField.Checked("serial", check_optional(str), check_optional(str))
    id_path = ObjectField.Checked("id_path", check_optional(str), check_optional(str))
    size = ObjectField.Checked("size", check(int), check(int))
    block_size = ObjectField.Checked("block_size", check(int), check(int))
    uuid = ObjectField.Checked("uuid", check(str), check(str))
    tags = ObjectField.Checked("tags", check(list), check(list))

    available_size = ObjectField.Checked("available_size", check(int), readonly=True)
    used_size = ObjectField.Checked("used_size", check(int), readonly=True)
    used_for = ObjectField.Checked("used_for", check(str), readonly=True)
    partition_table_type = ObjectField.Checked(
        "partition_table_type", to(PartitionTableType), readonly=True
    )

    partitions = ObjectFieldRelatedSet("partitions", "Partitions")
    filesystem = ObjectFieldRelated("filesystem", "Filesystem", readonly=True)

    def __repr__(self):
        if self.type == BlockDeviceType.PHYSICAL:
            return super(BlockDevice, self).__repr__(
                name="PhysicalBlockDevice",
                fields={"name", "model", "serial", "id_path"},
            )
        elif self.type == BlockDeviceType.VIRTUAL:
            return super(BlockDevice, self).__repr__(
                name="VirtualBlockDevice", fields={"name"}
            )
        else:
            raise ValueError("Unknown type: %s" % self.type)

    async def save(self):
        """Save this block device."""
        old_tags = list(self._orig_data["tags"])
        new_tags = list(self.tags)
        self._changed_data.pop("tags", None)
        await super(BlockDevice, self).save()
        for tag_name in new_tags:
            if tag_name not in old_tags:
                await self._handler.add_tag(
                    system_id=self.node.system_id, id=self.id, tag=tag_name
                )
            else:
                old_tags.remove(tag_name)
        for tag_name in old_tags:
            await self._handler.remove_tag(
                system_id=self.node.system_id, id=self.id, tag=tag_name
            )
        self._orig_data["tags"] = new_tags
        self._data["tags"] = list(new_tags)

    async def delete(self):
        """Delete this block device."""
        await self._handler.delete(system_id=self.node.system_id, id=self.id)

    async def set_as_boot_disk(self):
        """Set as boot disk for this node."""
        await self._handler.set_boot_disk(system_id=self.node.system_id, id=self.id)

    async def format(self, fstype, *, uuid=None):
        """Format this block device."""
        self._reset(
            await self._handler.format(
                system_id=self.node.system_id, id=self.id, fstype=fstype, uuid=uuid
            )
        )

    async def unformat(self):
        """Unformat this block device."""
        self._reset(
            await self._handler.unformat(system_id=self.node.system_id, id=self.id)
        )

    async def mount(self, mount_point, *, mount_options=None):
        """Mount this block device."""
        self._reset(
            await self._handler.mount(
                system_id=self.node.system_id,
                id=self.id,
                mount_point=mount_point,
                mount_options=mount_options,
            )
        )

    async def unmount(self):
        """Unmount this block device."""
        self._reset(
            await self._handler.unmount(system_id=self.node.system_id, id=self.id)
        )


class BlockDevicesType(ObjectType):
    """Metaclass for `BlockDevices`."""

    async def read(cls, node):
        """Get list of `BlockDevice`'s for `node`."""
        if isinstance(node, str):
            system_id = node
        elif isinstance(node, Node):
            system_id = node.system_id
        else:
            raise TypeError("node must be a Node or str, not %s" % type(node).__name__)
        data = await cls._handler.read(system_id=system_id)
        return cls(
            cls._object(item, local_data={"node_system_id": system_id}) for item in data
        )

    async def create(
        cls,
        node: Union[Node, str],
        name: str,
        *,
        model: str = None,
        serial: str = None,
        id_path: str = None,
        size: int = None,
        block_size: int = 512,
        tags: Iterable[str] = None
    ):
        """
        Create a physical block device on a Node.

        Either model and serial or id_path must be provided when creating a
        `BlockDevice`. Size (bytes) is always required.

        NOTE: It is recommended to use the MAAS commissioning process to
        discover `BlockDevice`'s on a machine. Getting any of this information
        incorrect can result on the machine failing to deploy.

        :param node: Node to create the block device on.
        :type node: `Node` or `str`
        :param name: The name for the block device.
        :type name: `str`
        :param model: The model number for the block device.
        :type model: `str`
        :param serial: The serial number for the block device.
        :type serial: `str`
        :param id_path: Unique path that identifies the device no matter
            the kernel the machine boots. Use only when the block device
            does not have a model and serial number.
        :type id_path: `str`
        :param size: The size of the block device in bytes.
        :type size: `int`
        :param block_size: The block size of the block device in bytes.
        :type block_size: `int`
        :param tags: List of tags to add to the block device.
        :type tags: sequence of `str`
        """
        params = {}
        if isinstance(node, str):
            params["system_id"] = node
        elif isinstance(node, Node):
            params["system_id"] = node.system_id
        else:
            raise TypeError(
                "node must be a Node or str, not %s" % (type(node).__name__)
            )

        if not size or size < 0:
            raise ValueError("size must be provided and greater than zero.")
        if not block_size or block_size < 0:
            raise ValueError("block_size must be provided and greater than zero.")
        if model and not serial:
            raise ValueError("serial must be provided when model is provided.")
        if not model and serial:
            raise ValueError("model must be provided when serial is provided.")
        if not model and not serial and not id_path:
            raise ValueError(
                "Either model/serial is provided or id_path must be provided."
            )

        params.update(
            remove_None(
                {
                    "name": name,
                    "model": model,
                    "serial": serial,
                    "id_path": id_path,
                    "size": size,
                    "block_size": block_size,
                }
            )
        )
        device = cls._object(await cls._handler.create(**params))
        if tags:
            device.tags = tags
            await device.save()
        return device


class BlockDevices(ObjectSet, metaclass=BlockDevicesType):
    """The set of block devices on a machine."""

    @property
    def by_name(self):
        """Return mapping of name of block device to `BlockDevice`."""
        return {bd.name: bd for bd in self}

    def get_by_name(self, name):
        """Return a `BlockDevice` by its name."""
        return self.by_name[name]



# Copyright 2016 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handling of MAAS API credentials.

The API client deals with credentials consisting of 3 elements: consumer key,
token key, and token secret. These are parts of the OAuth 1.0 specification.
The consumer secret is hard-wired to the empty string.

Credentials are represented internally as `Credentials` tuples, but can also
be converted to a colon-separated string format for easy transport between
processes.
"""

__all__ = ["Credentials"]

from collections import namedtuple
import typing


CredentialsBase = namedtuple(
    "CredentialsBase", ("consumer_key", "token_key", "token_secret")
)


class Credentials(CredentialsBase):
    """MAAS API Credentials."""

    __slots__ = ()

    @classmethod
    def parse(cls, credentials) -> typing.Optional["Credentials"]:
        """Parse/interpret some given credentials.

        These may take the form of:

        * An empty string.

        * An empty sequence.

        * A string, containing three parts (consumer key, token key, and token
          secret) separated by colons.

        * A sequence of three strings (consumer key, token key, and token
          secret).

        * None.

        """
        if credentials is None:
            return None
        elif isinstance(credentials, cls):
            return credentials
        elif isinstance(credentials, str):
            if credentials == "":
                return None
            elif credentials.count(":") == 2:
                return cls(*credentials.split(":"))
            else:
                raise ValueError(
                    "Malformed credentials. Expected 3 colon-separated "
                    "parts, got %r." % (credentials,)
                )
        else:
            parts = list(credentials)
            if len(parts) == 0:
                return None
            elif len(parts) == 3:
                return cls(*parts)
            else:
                raise ValueError(
                    "Malformed credentials. Expected 3 parts, "
                    "got %r." % (credentials,)
                )

    def __str__(self):
        return ":".join(self)



"""Tests for MAAS configuration and suchlike."""

from testtools.matchers import HasLength

from .. import maas
from ...testing import TestCase


def find_getters(cls):
    return {
        name[4:]: getattr(cls, name)
        for name in dir(cls)
        if name.startswith("get_") and name != "get_config"
    }


def find_setters(cls):
    return {
        name[4:]: getattr(cls, name)
        for name in dir(cls)
        if name.startswith("set_") and name != "set_config"
    }


class TestConfiguration(TestCase):
    def test__every_getter_has_a_setter_and_vice_versa(self):
        getters, setters = find_getters(maas.MAAS), find_setters(maas.MAAS)
        getters_without_setters = set(getters).difference(setters)
        self.assertThat(getters_without_setters, HasLength(0))
        setters_without_getters = set(setters).difference(getters)
        self.assertThat(setters_without_getters, HasLength(0))



"""Test for `maas.client.viscera.spaces`."""

import random

from testtools.matchers import Equals

from ..spaces import DeleteDefaultSpace, Space, Spaces

from ..testing import bind
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Spaces and Space. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(Spaces, Space)


class TestSpaces(TestCase):
    def test__spaces_create(self):
        Spaces = make_origin().Spaces
        name = make_string_without_spaces()
        description = make_string_without_spaces()
        Spaces._handler.create.return_value = {
            "id": 1,
            "name": name,
            "description": description,
        }
        Spaces.create(name=name, description=description)
        Spaces._handler.create.assert_called_once_with(
            name=name, description=description
        )

    def test__spaces_read(self):
        """Spaces.read() returns a list of Spaces."""
        Spaces = make_origin().Spaces
        spaces = [
            {"id": random.randint(0, 100), "name": make_string_without_spaces()}
            for _ in range(3)
        ]
        Spaces._handler.read.return_value = spaces
        spaces = Spaces.read()
        self.assertThat(len(spaces), Equals(3))


class TestSpace(TestCase):
    def test__space_get_default(self):
        Space = make_origin().Space
        Space._handler.read.return_value = {
            "id": 0,
            "name": make_string_without_spaces(),
        }
        Space.get_default()
        Space._handler.read.assert_called_once_with(id=0)

    def test__space_read(self):
        Space = make_origin().Space
        space = {"id": random.randint(0, 100), "name": make_string_without_spaces()}
        Space._handler.read.return_value = space
        self.assertThat(Space.read(id=space["id"]), Equals(Space(space)))
        Space._handler.read.assert_called_once_with(id=space["id"])

    def test__space_delete(self):
        Space = make_origin().Space
        space_id = random.randint(1, 100)
        space = Space({"id": space_id, "name": make_string_without_spaces()})
        space.delete()
        Space._handler.delete.assert_called_once_with(id=space_id)

    def test__space_delete_default(self):
        Space = make_origin().Space
        space = Space({"id": 0, "name": make_string_without_spaces()})
        self.assertRaises(DeleteDefaultSpace, space.delete)



"""Test for `maas.client.viscera.subnets`."""

import random

from testtools.matchers import Equals, IsInstance, MatchesAll, MatchesStructure

from ..subnets import Subnet, Subnets

from ..vlans import Vlan, Vlans

from ..testing import bind
from ...enum import RDNSMode
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with Subnets, Subnet, Vlans, Vlan.
    """
    return bind(Subnets, Subnet, Vlans, Vlan)


class TestSubnets(TestCase):
    def test__subnets_create(self):
        Subnets = make_origin().Subnets
        cidr = make_string_without_spaces()
        vlan = random.randint(5000, 8000)
        name = make_string_without_spaces()
        description = make_string_without_spaces()
        Subnets._handler.create.return_value = {
            "id": 1,
            "cidr": cidr,
            "vlan": vlan,
            "name": name,
            "description": description,
        }
        Subnets.create(cidr, vlan, name=name, description=description)
        Subnets._handler.create.assert_called_once_with(
            cidr=cidr, vlan=vlan, name=name, description=description
        )

    def test__subnets_read(self):
        """Subnets.read() returns a list of Subnets."""
        Subnets = make_origin().Subnets
        subnets = [
            {"id": random.randint(0, 100), "name": make_string_without_spaces()}
            for _ in range(3)
        ]
        Subnets._handler.read.return_value = subnets
        subnets = Subnets.read()
        self.assertThat(len(subnets), Equals(3))


class TestSubnet(TestCase):
    def test__subnet_read(self):
        Subnet = make_origin().Subnet
        vlan_id = random.randint(5000, 8000)
        subnet = {
            "id": random.randint(0, 100),
            "name": make_string_without_spaces(),
            "vlan": {"id": vlan_id},
            "rdns_mode": 2,
        }
        Subnet._handler.read.return_value = subnet
        self.assertThat(Subnet.read(id=subnet["id"]), Equals(Subnet(subnet)))
        self.assertThat(
            Subnet(subnet).vlan,
            MatchesAll(IsInstance(Vlan), MatchesStructure.byEquality(id=vlan_id)),
        )
        self.assertThat(Subnet(subnet).rdns_mode, Equals(RDNSMode.RFC2317))
        Subnet._handler.read.assert_called_once_with(id=subnet["id"])

    def test__subnet_update_vlan(self):
        origin = make_origin()
        Subnet, Vlan = origin.Subnet, origin.Vlan
        Subnet._handler.params = ["id"]
        subnet_id = random.randint(1, 100)
        subnet = Subnet(
            {
                "id": subnet_id,
                "name": make_string_without_spaces(),
                "vlan": {"id": random.randint(1, 100)},
            }
        )
        new_vlan = Vlan(
            {"id": random.randint(101, 200), "fabric_id": random.randint(101, 200)}
        )
        subnet.vlan = new_vlan
        Subnet._handler.update.return_value = {
            "id": subnet.id,
            "name": subnet.name,
            "vlan": {"id": new_vlan.id},
        }
        subnet.save()
        Subnet._handler.update.assert_called_once_with(id=subnet_id, vlan=new_vlan.id)

    def test__subnet_doesnt_update_vlan_if_same(self):
        Subnet = make_origin().Subnet
        subnet_id = random.randint(1, 100)
        subnet = Subnet(
            {
                "id": subnet_id,
                "name": make_string_without_spaces(),
                "vlan": {"id": random.randint(1, 100)},
            }
        )
        subnet.vlan = subnet.vlan
        subnet.save()
        self.assertEqual(0, Subnet._handler.update.call_count)

    def test__subnet_delete(self):
        Subnet = make_origin().Subnet
        subnet_id = random.randint(1, 100)
        subnet = Subnet({"id": subnet_id, "name": make_string_without_spaces()})
        subnet.delete()
        Subnet._handler.delete.assert_called_once_with(id=subnet_id)



"""Commands for running interactive and non-interactive shells."""

__all__ = ["register"]

import code
import sys
import textwrap
import tokenize

from . import colorized, Command, PROFILE_DEFAULT, PROFILE_NAMES
from .. import bones, facade, viscera


class cmd_shell(Command):
    """Start a python shell to interact with python-libmaas.

    If IPython is available it will be used, otherwise the familiar Python
    REPL will be started. If a script is piped in, it is read in its entirety
    then executed with the same namespace as the interactive shell.
    """

    profile_name_choices = PROFILE_NAMES
    profile_name_default = None if PROFILE_DEFAULT is None else PROFILE_DEFAULT.name

    def __init__(self, parser):
        super(cmd_shell, self).__init__(parser)
        if len(self.profile_name_choices) == 0:
            # There are no profiles, but we still offer the --profile-name
            # option so that users get a useful "profile not found" error
            # message instead of something more cryptic. Note that the help
            # string differs too.
            parser.add_argument(
                "--profile-name",
                metavar="NAME",
                required=False,
                default=None,
                help=(
                    "The name of the remote MAAS instance to use. "
                    "No profiles are currently defined; use the `profiles` "
                    "command to create one."
                ),
            )
        else:
            parser.add_argument(
                "--profile-name",
                metavar="NAME",
                required=False,
                choices=self.profile_name_choices,
                default=self.profile_name_default,
                help=(
                    "The name of the remote MAAS instance to use."
                    + (
                        ""
                        if self.profile_name_default is None
                        else " [default: %(default)s]"
                    )
                ),
            )
        parser.add_argument(
            "--viscera",
            action="store_true",
            default=False,
            help=(
                "Create a pre-canned viscera `Origin` for the selected "
                "profile. This is available as `origin` in the shell's "
                "namespace. You probably do not need this unless you're "
                "developing python-libmaas itself."
            ),
        )
        parser.add_argument(
            "--bones",
            action="store_true",
            default=False,
            help=(
                "Create a pre-canned bones `Session` for the selected "
                "profile. This is available as `session` in the shell's "
                "namespace. You probably do not need this unless you're "
                "developing python-libmaas itself."
            ),
        )
        parser.add_argument(
            "script",
            metavar="SCRIPT",
            nargs="?",
            default=None,
            help=(
                "Python script to run in the shell's namespace. An "
                "interactive shell is started if none is given."
            ),
        )

    def __call__(self, options):
        """Execute this command."""

        # The namespace that code will run in.
        namespace = {}
        # Descriptions of the namespace variables.
        descriptions = {}

        # If a profile has been selected, set up a `bones` session and a
        # `viscera` origin in the default namespace.
        if options.profile_name is not None:
            session = bones.SessionAPI.fromProfileName(options.profile_name)
            if options.bones:
                namespace["session"] = session
                descriptions["session"] = (
                    "A pre-canned `bones` session for '%s'." % options.profile_name
                )
            origin = viscera.Origin(session)
            if options.viscera:
                namespace["origin"] = origin
                descriptions["origin"] = (
                    "A pre-canned `viscera` origin for '%s'." % options.profile_name
                )
            client = facade.Client(origin)
            namespace["client"] = client
            descriptions["client"] = (
                "A pre-canned client for '%s'." % options.profile_name
            )

        if options.script is None:
            if sys.stdin.isatty() and sys.stdout.isatty():
                self._run_interactive(namespace, descriptions)
            else:
                self._run_stdin(namespace)
        else:
            self._run_script(namespace, options.script)

    @staticmethod
    def _run_interactive(namespace, descriptions):
        # We at a fully interactive terminal — i.e. stdin AND stdout are
        # connected to the TTY — so display some introductory text...
        banner = ["{automagenta}Welcome to the MAAS shell.{/automagenta}"]
        if len(descriptions) > 0:
            banner += ["", "Predefined objects:", ""]
            wrap = textwrap.TextWrapper(60, "    ", "    ").wrap
            sortkey = lambda name: (name.casefold(), name)
            for name in sorted(descriptions, key=sortkey):
                banner.append("  {autoyellow}%s{/autoyellow}:" % name)
                banner.extend(wrap(descriptions[name]))
                banner.append("")
        for line in banner:
            print(colorized(line))
        # ... then start IPython, or the plain familiar Python REPL if
        # IPython is not installed.
        try:
            import IPython
        except ImportError:
            code.InteractiveConsole(namespace).interact(" ")
        else:
            IPython.start_ipython(argv=[], display_banner=False, user_ns=namespace)

    @staticmethod
    def _run_script(namespace, filename):
        namespace = dict(namespace, __file__=filename)
        with tokenize.open(filename) as fd:
            code = compile(fd.read(), filename, "exec")
        exec(code, namespace, namespace)

    @staticmethod
    def _run_stdin(namespace):
        namespace = dict(namespace, __file__="<stdin>")
        code = compile(sys.stdin.read(), "<stdin>", "exec")
        exec(code, namespace, namespace)


def register(parser):
    """Register profile commands with the given parser."""
    cmd_shell.register(parser)
